#!/usr/bin/env python3

import hashlib
import os
import pickle
import re
import string
from collections import defaultdict
from copy import deepcopy
from itertools import product
from pathlib import Path

import ruamel.yaml
from osaca import __version__, utils
from osaca.parser import ParserX86ATT
from ruamel.yaml.compat import StringIO


class MachineModel(object):
    WILDCARD = "*"
    INTERNAL_VERSION = 1  # increase whenever self._data format changes to invalidate cache!
    _runtime_cache = {}

    def __init__(self, arch=None, path_to_yaml=None, isa=None, lazy=False):
        if not arch and not path_to_yaml:
            if not isa:
                raise ValueError("One of arch, path_to_yaml and isa must be specified")
            self._data = {
                "osaca_version": str(__version__),
                "micro_architecture": None,
                "arch_code": None,
                "isa": isa,
                "ROB_size": None,
                "retired_uOps_per_cycle": None,
                "scheduler_size": None,
                "hidden_loads": None,
                "load_latency": {},
                "load_throughput": [
                    {
                        "base": b,
                        "index": i,
                        "offset": o,
                        "scale": s,
                        "port_pressure": [],
                    }
                    for b, i, o, s in product(["gpr"], ["gpr", None], ["imd", None], [1, 8])
                ],
                "load_throughput_default": [],
                "store_throughput": [],
                "store_throughput_default": [],
                "store_to_load_forward_latency": None,
                "ports": [],
                "port_model_scheme": None,
                "instruction_forms": [],
                "instruction_forms_dict": defaultdict(list),
            }
        else:
            if arch and path_to_yaml:
                raise ValueError("Only one of arch and path_to_yaml is allowed.")
            self._path = path_to_yaml
            self._arch = arch
            if arch:
                self._arch = arch.lower()
                self._path = utils.find_datafile(self._arch + ".yml")
            # Check runtime cache
            if self._path in MachineModel._runtime_cache and not lazy:
                self._data = MachineModel._runtime_cache[self._path]
            # check if file is cached
            cached = self._get_cached(self._path) if not lazy else False
            if cached:
                self._data = cached
            else:
                yaml = self._create_yaml_object()
                # otherwise load
                with open(self._path, "r") as f:
                    if not lazy:
                        self._data = yaml.load(f)
                    else:
                        file_content = ""
                        line = f.readline()
                        while "instruction_forms:" not in line:
                            file_content += line
                            line = f.readline()
                        self._data = yaml.load(file_content)
                        self._data["instruction_forms"] = []
                # separate multi-alias instruction forms
                for entry in [
                    x for x in self._data["instruction_forms"] if isinstance(x["name"], list)
                ]:
                    for name in entry["name"]:
                        new_entry = {"name": name}
                        for k in [x for x in entry.keys() if x != "name"]:
                            new_entry[k] = entry[k]
                        self._data["instruction_forms"].append(new_entry)
                    # remove old entry
                    self._data["instruction_forms"].remove(entry)
                # Normalize instruction_form names (to UPPERCASE) and build dict for faster access:
                self._data["instruction_forms_dict"] = defaultdict(list)
                for iform in self._data["instruction_forms"]:
                    iform["name"] = iform["name"].upper()
                    self._data["instruction_forms_dict"][iform["name"]].append(iform)
                self._data["internal_version"] = self.INTERNAL_VERSION

                if not lazy:
                    # cache internal representation for future use
                    self._write_in_cache(self._path)
            # Store in runtime cache
            if not lazy:
                MachineModel._runtime_cache[self._path] = self._data

    def get(self, key, default=None):
        """Return config entry for key or default/None."""
        return self._data.get(key, default)

    def __getitem__(self, key):
        """Return configuration entry."""
        return self._data[key]

    def __contains__(self, key):
        """Return true if configuration key is present."""
        return key in self._data

    ######################################################

    def get_instruction(self, name, operands):
        """Find and return instruction data from name and operands."""
        # For use with dict instead of list as DB
        if name is None:
            return None
        name_matched_iforms = self._data["instruction_forms_dict"].get(name.upper(), [])
        try:
            return next(
                instruction_form
                for instruction_form in name_matched_iforms
                if self._match_operands(
                    instruction_form["operands"] if "operands" in instruction_form else [],
                    operands,
                )
            )
        except StopIteration:
            return None
        except TypeError as e:
            print("\nname: {}\noperands: {}".format(name, operands))
            raise TypeError from e

    def average_port_pressure(self, port_pressure):
        """Construct average port pressure list from instruction data."""
        port_list = self._data["ports"]
        average_pressure = [0.0] * len(port_list)
        for cycles, ports in port_pressure:
            for p in ports:
                try:
                    average_pressure[port_list.index(p)] += cycles / len(ports)
                except ValueError as e:
                    raise KeyError("Port {!r} not in port list.".format(p)) from e
        return average_pressure

    def set_instruction(
        self,
        name,
        operands=None,
        latency=None,
        port_pressure=None,
        throughput=None,
        uops=None,
    ):
        """Import instruction form information."""
        # If it already exists. Overwrite information.
        instr_data = self.get_instruction(name, operands)
        if instr_data is None:
            instr_data = {}
            self._data["instruction_forms"].append(instr_data)
            self._data["instruction_forms_dict"][name].append(instr_data)

        instr_data["name"] = name
        instr_data["operands"] = operands
        instr_data["latency"] = latency
        instr_data["port_pressure"] = port_pressure
        instr_data["throughput"] = throughput
        instr_data["uops"] = uops

    def set_instruction_entry(self, entry):
        """Import instruction as entry object form information."""
        self.set_instruction(
            entry["name"],
            entry["operands"] if "operands" in entry else None,
            entry["latency"] if "latency" in entry else None,
            entry["port_pressure"] if "port_pressure" in entry else None,
            entry["throughput"] if "throughput" in entry else None,
            entry["uops"] if "uops" in entry else None,
        )

    def add_port(self, port):
        """Add port in port model of current machine model."""
        if port not in self._data["ports"]:
            self._data["ports"].append(port)

    def get_ISA(self):
        """Return ISA of :class:`MachineModel`."""
        return self._data["isa"].lower()

    def get_arch(self):
        """Return micro-architecture code of :class:`MachineModel`."""
        return self._data["arch_code"].lower()

    def get_ports(self):
        """Return port model of :class:`MachineModel`."""
        return self._data["ports"]

    def has_hidden_loads(self):
        """Return if model has hidden loads."""
        if "hidden_loads" in self._data:
            return self._data["hidden_loads"]
        return False

    def get_load_latency(self, reg_type):
        """Return load latency for given register type."""
        return self._data["load_latency"][reg_type] if self._data["load_latency"][reg_type] else 0

    def get_load_throughput(self, memory):
        """Return load thorughput for given register type."""
        ld_tp = [m for m in self._data["load_throughput"] if self._match_mem_entries(memory, m)]
        if len(ld_tp) > 0:
            return ld_tp[0]["port_pressure"].copy()
        return self._data["load_throughput_default"].copy()

    def get_store_latency(self, reg_type):
        """Return store latency for given register type."""
        # assume 0 for now, since load-store-dependencies currently not detectable
        return 0

    def get_store_throughput(self, memory):
        """Return store throughput for given register type."""
        st_tp = [m for m in self._data["store_throughput"] if self._match_mem_entries(memory, m)]
        if len(st_tp) > 0:
            return st_tp[0]["port_pressure"].copy()
        return self._data["store_throughput_default"].copy()

    def _match_mem_entries(self, mem, i_mem):
        """Check if memory addressing ``mem`` and ``i_mem`` are of the same type."""
        if self._data["isa"].lower() == "aarch64":
            return self._is_AArch64_mem_type(i_mem, mem)
        if self._data["isa"].lower() == "x86":
            return self._is_x86_mem_type(i_mem, mem)

    def get_data_ports(self):
        """Return all data ports (i.e., ports with D-suffix) of current model."""
        data_port = re.compile(r"^[0-9]+D$")
        data_ports = [x for x in filter(data_port.match, self._data["ports"])]
        return data_ports

    @staticmethod
    def get_full_instruction_name(instruction_form):
        """Get one instruction name string including the mnemonic and all operands."""
        operands = []
        for op in instruction_form["operands"]:
            op_attrs = [
                y + ":" + str(op[y])
                for y in list(filter(lambda x: True if x != "class" else False, op))
            ]
            operands.append("{}({})".format(op["class"], ",".join(op_attrs)))
        return "{}  {}".format(instruction_form["name"].lower(), ",".join(operands))

    @staticmethod
    def get_isa_for_arch(arch):
        """Return ISA for given micro-arch ``arch``."""
        arch_dict = {
            "a64fx": "aarch64",
            "tx2": "aarch64",
            "n1": "aarch64",
            "zen1": "x86",
            "zen+": "x86",
            "zen2": "x86",
            "con": "x86",  # Intel Conroe
            "wol": "x86",  # Intel Wolfdale
            "snb": "x86",
            "ivb": "x86",
            "hsw": "x86",
            "bdw": "x86",
            "skl": "x86",
            "skx": "x86",
            "csx": "x86",
            "wsm": "x86",
            "nhm": "x86",
            "kbl": "x86",
            "cnl": "x86",
            "cfl": "x86",
            "icl": "x86",
        }
        arch = arch.lower()
        if arch in arch_dict:
            return arch_dict[arch].lower()
        else:
            raise ValueError("Unknown architecture {!r}.".format(arch))

    def dump(self, stream=None):
        """Dump machine model to stream or return it as a ``str`` if no stream is given."""
        # Replace instruction form's port_pressure with styled version for RoundtripDumper
        formatted_instruction_forms = deepcopy(self._data["instruction_forms"])
        for instruction_form in formatted_instruction_forms:
            if instruction_form["port_pressure"] is not None:
                cs = ruamel.yaml.comments.CommentedSeq(instruction_form["port_pressure"])
                cs.fa.set_flow_style()
                instruction_form["port_pressure"] = cs

        # Replace load_throughput with styled version for RoundtripDumper
        formatted_load_throughput = []
        for lt in self._data["load_throughput"]:
            cm = ruamel.yaml.comments.CommentedMap(lt)
            cm.fa.set_flow_style()
            formatted_load_throughput.append(cm)

        # Create YAML object
        yaml = self._create_yaml_object()
        if not stream:
            stream = StringIO()

        yaml.dump(
            {
                k: v
                for k, v in self._data.items()
                if k
                not in [
                    "instruction_forms",
                    "instruction_forms_dict",
                    "load_throughput",
                    "internal_version",
                ]
            },
            stream,
        )
        yaml.dump({"load_throughput": formatted_load_throughput}, stream)
        yaml.dump({"instruction_forms": formatted_instruction_forms}, stream)

        if isinstance(stream, StringIO):
            return stream.getvalue()

    ######################################################

    def _get_cached(self, filepath):
        """
        Check if machine model is cached and if so, load it.

        :param filepath: path to check for cached machine model
        :type filepath: str
        :returns: cached DB if existing, `False` otherwise
        """
        p = Path(filepath)
        hexhash = hashlib.sha256(p.read_bytes()).hexdigest()

        # 1. companion cachefile: same location, with '.<name>_<sha512hash>.pickle'
        companion_cachefile = p.with_name("." + p.stem + "_" + hexhash).with_suffix(".pickle")
        if companion_cachefile.exists():
            # companion file (must be up-to-date, due to equal hash)
            with companion_cachefile.open("rb") as f:
                data = pickle.load(f)
            if data.get("internal_version") == self.INTERNAL_VERSION:
                return data

        # 2. home cachefile: ~/.osaca/cache/<name>_<sha512hash>.pickle
        home_cachefile = (Path(utils.CACHE_DIR) / (p.stem + "_" + hexhash)).with_suffix(".pickle")
        if home_cachefile.exists():
            # home file (must be up-to-date, due to equal hash)
            with home_cachefile.open("rb") as f:
                data = pickle.load(f)
            if data.get("internal_version") == self.INTERNAL_VERSION:
                return data
        return False

    def _write_in_cache(self, filepath):
        """
        Write machine model to cache

        :param filepath: path to store DB
        :type filepath: str
        """
        p = Path(filepath)
        hexhash = hashlib.sha256(p.read_bytes()).hexdigest()
        # 1. companion cachefile: same location, with '.<name>_<sha512hash>.pickle'
        companion_cachefile = p.with_name("." + p.stem + "_" + hexhash).with_suffix(".pickle")
        if os.access(str(companion_cachefile.parent), os.W_OK):
            with companion_cachefile.open("wb") as f:
                pickle.dump(self._data, f)
                return

        # 2. home cachefile: ~/.osaca/cache/<name>_<sha512hash>.pickle
        cache_dir = Path(utils.CACHE_DIR)
        try:
            os.makedirs(cache_dir, exist_ok=True)
        except OSError:
            return
        home_cachefile = (cache_dir / (p.stem + "_" + hexhash)).with_suffix(".pickle")
        if os.access(str(home_cachefile.parent), os.W_OK):
            with home_cachefile.open("wb") as f:
                pickle.dump(self._data, f)

    def _get_key(self, name, operands):
        """Get unique instruction form key for dict DB."""
        key_string = name.lower() + "-"
        if operands is None:
            return key_string[:-1]
        key_string += "_".join([self._get_operand_hash(op) for op in operands])
        return key_string

    def _get_operand_hash(self, operand):
        """Get unique key for operand for dict DB"""
        operand_string = ""
        if "class" in operand:
            # DB entry
            opclass = operand["class"]
        else:
            # parsed instruction
            opclass = list(operand.keys())[0]
            operand = operand[opclass]
        if opclass == "immediate":
            # Immediate
            operand_string += "i"
        elif opclass == "register":
            # Register
            if "prefix" in operand:
                operand_string += operand["prefix"]
                operand_string += operand["shape"] if "shape" in operand else ""
            elif "name" in operand:
                operand_string += "r" if operand["name"] == "gpr" else operand["name"][0]
        elif opclass == "memory":
            # Memory
            operand_string += "m"
            operand_string += "b" if operand["base"] is not None else ""
            operand_string += "o" if operand["offset"] is not None else ""
            operand_string += "i" if operand["index"] is not None else ""
            operand_string += (
                "s" if operand["scale"] == self.WILDCARD or operand["scale"] > 1 else ""
            )
            if "pre-indexed" in operand:
                operand_string += "r" if operand["pre-indexed"] else ""
                operand_string += "p" if operand["post-indexed"] else ""
        return operand_string

    def _create_db_operand_aarch64(self, operand):
        """Create instruction form operand for DB out of operand string."""
        if operand == "i":
            return {"class": "immediate", "imd": "int"}
        elif operand in "wxbhsdq":
            return {"class": "register", "prefix": operand}
        elif operand.startswith("v"):
            return {"class": "register", "prefix": "v", "shape": operand[1:2]}
        elif operand.startswith("m"):
            return {
                "class": "memory",
                "base": "x" if "b" in operand else None,
                "offset": "imd" if "o" in operand else None,
                "index": "gpr" if "i" in operand else None,
                "scale": 8 if "s" in operand else 1,
                "pre-indexed": True if "r" in operand else False,
                "post-indexed": True if "p" in operand else False,
            }
        else:
            raise ValueError("Parameter {} is not a valid operand code".format(operand))

    def _create_db_operand_x86(self, operand):
        """Create instruction form operand for DB out of operand string."""
        if operand == "r":
            return {"class": "register", "name": "gpr"}
        elif operand in "xyz":
            return {"class": "register", "name": operand + "mm"}
        elif operand == "i":
            return {"class": "immediate", "imd": "int"}
        elif operand.startswith("m"):
            return {
                "class": "memory",
                "base": "gpr" if "b" in operand else None,
                "offset": "imd" if "o" in operand else None,
                "index": "gpr" if "i" in operand else None,
                "scale": 8 if "s" in operand else 1,
            }
        else:
            raise ValueError("Parameter {} is not a valid operand code".format(operand))

    def _check_for_duplicate(self, name, operands):
        """
        Check if instruction form exists at least twice in DB.

        :param str name: mnemonic of instruction form
        :param list operands: instruction form operands

        :returns: `True`, if duplicate exists, `False` otherwise
        """
        matches = [
            instruction_form
            for instruction_form in self._data["instruction_forms"]
            if instruction_form["name"].lower() == name.lower()
            and self._match_operands(instruction_form["operands"], operands)
        ]
        if len(matches) > 1:
            return True
        return False

    def _match_operands(self, i_operands, operands):
        """Check if all operand types of ``i_operands`` and ``operands`` match."""
        operands_ok = True
        if len(operands) != len(i_operands):
            return False
        for idx, operand in enumerate(operands):
            i_operand = i_operands[idx]
            operands_ok = operands_ok and self._check_operands(i_operand, operand)
        if operands_ok:
            return True
        else:
            return False

    def _check_operands(self, i_operand, operand):
        """Check if the types of operand ``i_operand`` and ``operand`` match."""
        # check for wildcard
        if self.WILDCARD in operand:
            if (
                "class" in i_operand
                and i_operand["class"] == "register"
                or "register" in i_operand
            ):
                return True
            else:
                return False
        if self._data["isa"].lower() == "aarch64":
            return self._check_AArch64_operands(i_operand, operand)
        if self._data["isa"].lower() == "x86":
            return self._check_x86_operands(i_operand, operand)

    def _check_AArch64_operands(self, i_operand, operand):
        """Check if the types of operand ``i_operand`` and ``operand`` match."""
        if "class" in operand:
            # compare two DB entries
            return self._compare_db_entries(i_operand, operand)
        # TODO support class wildcards
        # register
        if "register" in operand:
            if i_operand["class"] != "register":
                return False
            return self._is_AArch64_reg_type(i_operand, operand["register"])
        # memory
        if "memory" in operand:
            if i_operand["class"] != "memory":
                return False
            return self._is_AArch64_mem_type(i_operand, operand["memory"])
        # immediate
        if i_operand["class"] == "immediate" and i_operand["imd"] == self.WILDCARD:
            return "value" in operand or (
                "immediate" in operand and "value" in operand["immediate"]
            )
        if i_operand["class"] == "immediate" and i_operand["imd"] == "int":
            return ("value" in operand and operand.get("type", None) == "int") or (
                "immediate" in operand
                and "value" in operand["immediate"]
                and operand["immediate"].get("type", None) == "int"
            )
        if i_operand["class"] == "immediate" and i_operand["imd"] == "float":
            return ("float" in operand and operand.get("type", None) == "float") or (
                "immediate" in operand
                and "float" in operand["immediate"]
                and operand["immediate"].get("type", None) == "float"
            )
        if i_operand["class"] == "immediate" and i_operand["imd"] == "double":
            return ("double" in operand and operand.get("type", None) == "double") or (
                "immediate" in operand
                and "double" in operand["immediate"]
                and operand["immediate"].get("type", None) == "double"
            )
        # identifier
        if "identifier" in operand or (
            "immediate" in operand and "identifier" in operand["immediate"]
        ):
            return i_operand["class"] == "identifier"
        # prefetch option
        if "prfop" in operand:
            return i_operand["class"] == "prfop"
        # no match
        return False

    def _check_x86_operands(self, i_operand, operand):
        """Check if the types of operand ``i_operand`` and ``operand`` match."""
        if "class" in operand:
            # compare two DB entries
            return self._compare_db_entries(i_operand, operand)
        # register
        if "register" in operand:
            if i_operand["class"] != "register":
                return False
            return self._is_x86_reg_type(i_operand, operand["register"], consider_masking=False)
        # memory
        if "memory" in operand:
            if i_operand["class"] != "memory":
                return False
            return self._is_x86_mem_type(i_operand, operand["memory"])
        # immediate
        if "immediate" in operand or "value" in operand:
            return i_operand["class"] == "immediate" and i_operand["imd"] == "int"
        # identifier (e.g., labels)
        if "identifier" in operand:
            return i_operand["class"] == "identifier"

    def _compare_db_entries(self, operand_1, operand_2):
        """Check if operand types in DB format (i.e., not parsed) match."""
        operand_attributes = list(
            filter(
                lambda x: True if x != "source" and x != "destination" else False,
                operand_1,
            )
        )
        for key in operand_attributes:
            try:
                if operand_1[key] != operand_2[key] and not any(
                    [x == self.WILDCARD for x in [operand_1[key], operand_2[key]]]
                ):
                    return False
            except KeyError:
                return False
        return True

    def _is_AArch64_reg_type(self, i_reg, reg):
        """Check if register type match."""
        # check for wildcards
        if reg["prefix"] == self.WILDCARD or i_reg["prefix"] == self.WILDCARD:
            if "shape" in reg:
                if "shape" in i_reg and (
                    reg["shape"] == i_reg["shape"]
                    or self.WILDCARD in (reg["shape"] + i_reg["shape"])
                ):
                    return True
                return False
            return True
        # check for prefix and shape
        if reg["prefix"] != i_reg["prefix"]:
            return False
        if "shape" in reg:
            if "shape" in i_reg and (
                reg["shape"] == i_reg["shape"] or self.WILDCARD in (reg["shape"] + i_reg["shape"])
            ):
                return True
            return False
        return True

    def _is_x86_reg_type(self, i_reg, reg, consider_masking=False):
        """Check if register type match."""
        i_reg_name = i_reg["name"] if i_reg and "name" in i_reg else i_reg
        if reg is None:
            if i_reg is None:
                return True
            return False
        # check for wildcards
        if i_reg_name == self.WILDCARD or reg["name"] == self.WILDCARD:
            return True
        # differentiate between vector registers (mm, xmm, ymm, zmm) and others (gpr)
        parser_x86 = ParserX86ATT()
        if parser_x86.is_vector_register(reg):
            if reg["name"].rstrip(string.digits).lower() == i_reg_name:
                # Consider masking and zeroing for AVX512
                if consider_masking:
                    mask_ok = zero_ok = True
                    if "mask" in reg or "mask" in i_reg:
                        # one instruction is missing the masking while the other has it
                        mask_ok = False
                        # check for wildcard
                        if (
                            (
                                "mask" in reg
                                and reg["mask"].rstrip(string.digits).lower() == i_reg.get("mask")
                            )
                            or reg.get("mask") == self.WILDCARD
                            or i_reg.get("mask") == self.WILDCARD
                        ):
                            mask_ok = True
                        if bool("zeroing" in reg) ^ bool("zeroing" in i_reg):
                            # one instruction is missing zeroing while the other has it
                            zero_ok = False
                            # check for wildcard
                            if (
                                i_reg.get("zeroing") == self.WILDCARD
                                or reg.get("zeroing") == self.WILDCARD
                            ):
                                zero_ok = True
                        if not mask_ok or not zero_ok:
                            return False
                return True
        else:
            if i_reg_name == "gpr":
                return True
        return False

    def _is_AArch64_mem_type(self, i_mem, mem):
        """Check if memory addressing type match."""
        if (
            # check base
            (
                (mem["base"] is None and i_mem["base"] is None)
                or i_mem["base"] == self.WILDCARD
                or mem["base"]["prefix"] == i_mem["base"]
            )
            # check offset
            and (
                mem["offset"] == i_mem["offset"]
                or i_mem["offset"] == self.WILDCARD
                or (
                    mem["offset"] is not None
                    and "identifier" in mem["offset"]
                    and i_mem["offset"] == "identifier"
                )
                or (
                    mem["offset"] is not None
                    and "value" in mem["offset"]
                    and i_mem["offset"] == "imd"
                )
            )
            # check index
            and (
                mem["index"] == i_mem["index"]
                or i_mem["index"] == self.WILDCARD
                or (
                    mem["index"] is not None
                    and "prefix" in mem["index"]
                    and mem["index"]["prefix"] == i_mem["index"]
                )
            )
            # check scale
            and (
                mem["scale"] == i_mem["scale"]
                or i_mem["scale"] == self.WILDCARD
                or (mem["scale"] != 1 and i_mem["scale"] != 1)
            )
            # check pre-indexing
            and (
                i_mem["pre-indexed"] == self.WILDCARD
                or ("pre_indexed" in mem) == (i_mem["pre-indexed"])
            )
            # check post-indexing
            and (
                i_mem["post-indexed"] == self.WILDCARD
                or ("post_indexed" in mem) == (i_mem["post-indexed"])
            )
        ):
            return True
        return False

    def _is_x86_mem_type(self, i_mem, mem):
        """Check if memory addressing type match."""
        if (
            # check base
            (
                (mem["base"] is None and i_mem["base"] is None)
                or i_mem["base"] == self.WILDCARD
                or self._is_x86_reg_type(i_mem["base"], mem["base"])
            )
            # check offset
            and (
                mem["offset"] == i_mem["offset"]
                or i_mem["offset"] == self.WILDCARD
                or (
                    mem["offset"] is not None
                    and "identifier" in mem["offset"]
                    and i_mem["offset"] == "identifier"
                )
                or (
                    mem["offset"] is not None
                    and "value" in mem["offset"]
                    and (
                        i_mem["offset"] == "imd"
                        or (i_mem["offset"] is None and mem["offset"]["value"] == "0")
                    )
                )
                or (
                    mem["offset"] is not None
                    and "identifier" in mem["offset"]
                    and i_mem["offset"] == "id"
                )
            )
            # check index
            and (
                mem["index"] == i_mem["index"]
                or i_mem["index"] == self.WILDCARD
                or (
                    mem["index"] is not None
                    and "name" in mem["index"]
                    and self._is_x86_reg_type(i_mem["index"], mem["index"])
                )
            )
            # check scale
            and (
                mem["scale"] == i_mem["scale"]
                or i_mem["scale"] == self.WILDCARD
                or (mem["scale"] != 1 and i_mem["scale"] != 1)
            )
        ):
            return True
        return False

    def _create_yaml_object(self):
        """Create YAML object for parsing and dumping DB"""
        yaml_obj = ruamel.yaml.YAML()
        yaml_obj.representer.add_representer(type(None), self.__represent_none)
        yaml_obj.default_flow_style = None
        yaml_obj.width = 120
        yaml_obj.representer.ignore_aliases = lambda *args: True
        return yaml_obj

    def __represent_none(self, yaml_obj, data):
        """YAML representation for `None`"""
        return yaml_obj.represent_scalar(u"tag:yaml.org,2002:null", u"~")



#!/usr/bin/env python3
"""
Unit tests for base assembly parser
"""

import os
import unittest

from osaca.parser import AttrDict, BaseParser


class TestBaseParser(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        try:
            self.parser = BaseParser()
        except NotImplementedError:
            pass
        with open(self._find_file("triad_x86_iaca.s")) as f:
            self.triad_code = f.read()
        with open(self._find_file("triad_arm_iaca.s")) as f:
            self.triad_code_arm = f.read()
        with open(self._find_file("kernel_x86.s")) as f:
            self.x86_code = f.read()
        with open(self._find_file("kernel_aarch64.s")) as f:
            self.aarch64_code = f.read()

    ##################
    # Test
    ##################

    def test_parse_file(self):
        with self.assertRaises(NotImplementedError):
            self.parser.parse_file(self.triad_code)

    def test_parse_line(self):
        line_instruction = "\t\tlea       2(%rax,%rax), %ecx #12.9"
        with self.assertRaises(NotImplementedError):
            self.parser.parse_line(line_instruction)

    def test_parse_instruction(self):
        instr1 = "\t\tvcvtsi2ss %edx, %xmm2, %xmm2\t\t\t#12.27"
        with self.assertRaises(NotImplementedError):
            self.parser.parse_instruction(instr1)

    def test_register_funcs(self):
        reg_a1 = AttrDict({"name": "rax"})
        reg_a2 = AttrDict({"name": "eax"})
        register_string = "v1.2d"
        with self.assertRaises(NotImplementedError):
            self.parser.is_reg_dependend_of(reg_a1, reg_a2)
        with self.assertRaises(NotImplementedError):
            self.parser.parse_register(register_string)
        with self.assertRaises(NotImplementedError):
            self.parser.is_gpr(reg_a1)
        with self.assertRaises(NotImplementedError):
            self.parser.is_vector_register(reg_a1)
        with self.assertRaises(NotImplementedError):
            self.parser.process_operand(reg_a1)
        with self.assertRaises(NotImplementedError):
            self.parser.get_full_reg_name(reg_a1)

    def test_normalize_imd(self):
        imd_hex_1 = {"value": "0x4f"}
        with self.assertRaises(NotImplementedError):
            self.parser.normalize_imd(imd_hex_1)

    def test_detect_ISA(self):
        self.assertEqual(BaseParser.detect_ISA(self.triad_code), "x86")
        self.assertEqual(BaseParser.detect_ISA(self.triad_code_arm), "aarch64")
        self.assertEqual(BaseParser.detect_ISA(self.x86_code), "x86")
        self.assertEqual(BaseParser.detect_ISA(self.aarch64_code), "aarch64")

    ##################
    # Helper functions
    ##################

    @staticmethod
    def _find_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestBaseParser)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python3
"""Attribute Dictionary to access dictionary entries as attributes."""


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    @staticmethod
    def convert_dict(dictionary):
        """
        Convert given dictionary to `AttrDict`.

        :param dictionary: `dict` to be converted
        :type dictionary: `dict`
        :returns: `AttrDict` representation of ``dictionary``
        """
        if isinstance(dictionary, type(list())):
            return [AttrDict.convert_dict(x) for x in dictionary]
        if isinstance(dictionary, type(dict())):
            for key in list(dictionary.keys()):
                entry = dictionary[key]
                if isinstance(entry, type(dict())) or isinstance(entry, type(AttrDict())):
                    dictionary[key] = AttrDict.convert_dict(dictionary[key])
                if isinstance(entry, type(list())):
                    dictionary[key] = [AttrDict.convert_dict(x) for x in entry]
            return AttrDict(dictionary)
        return dictionary



#!/usr/bin/env python3
"""
Unit tests for DB interface
"""
import os
import unittest
from io import StringIO

import osaca.db_interface as dbi
from osaca.db_interface import sanity_check
from osaca.semantics import MachineModel


class TestDBInterface(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        sample_entry = {
            "name": "DoItRightAndDoItFast",
            "operands": [
                {
                    "class": "memory",
                    "offset": "imd",
                    "base": "gpr",
                    "index": "gpr",
                    "scale": 8,
                },
                {"class": "register", "name": "xmm"},
            ],
            "throughput": 1.25,
            "latency": 125,
            "uops": 6,
        }
        self.entry_csx = sample_entry.copy()
        self.entry_tx2 = sample_entry.copy()
        self.entry_zen1 = sample_entry.copy()

        # self.entry_csx['port_pressure'] = [1.25, 0, 1.25, 0.5, 0.5, 0.5, 0.5, 0, 1.25, 1.25, 0]
        self.entry_csx["port_pressure"] = [[5, "0156"], [1, "23"], [1, ["2D", "3D"]]]
        # self.entry_tx2['port_pressure'] = [2.5, 2.5, 0, 0, 0.5, 0.5]
        self.entry_tx2["port_pressure"] = [[5, "01"], [1, "45"]]
        del self.entry_tx2["operands"][1]["name"]
        self.entry_tx2["operands"][1]["prefix"] = "x"
        # self.entry_zen1['port_pressure'] = [1, 1, 1, 1, 0, 1, 0, 0, 0, 0.5, 1, 0.5, 1]
        self.entry_zen1["port_pressure"] = [
            [4, "0123"],
            [1, "4"],
            [1, "89"],
            [2, ["8D", "9D"]],
        ]

    ###########
    # Tests
    ###########

    def test_add_single_entry(self):
        mm_csx = MachineModel("csx")
        mm_tx2 = MachineModel("tx2")
        mm_zen1 = MachineModel("zen1")
        num_entries_csx = len(mm_csx["instruction_forms"])
        num_entries_tx2 = len(mm_tx2["instruction_forms"])
        num_entries_zen1 = len(mm_zen1["instruction_forms"])

        mm_csx.set_instruction_entry(self.entry_csx)
        mm_tx2.set_instruction_entry(self.entry_tx2)
        mm_zen1.set_instruction_entry({"name": "empty_operation"})

        num_entries_csx = len(mm_csx["instruction_forms"]) - num_entries_csx
        num_entries_tx2 = len(mm_tx2["instruction_forms"]) - num_entries_tx2
        num_entries_zen1 = len(mm_zen1["instruction_forms"]) - num_entries_zen1

        self.assertEqual(num_entries_csx, 1)
        self.assertEqual(num_entries_tx2, 1)
        self.assertEqual(num_entries_zen1, 1)

    def test_invalid_add(self):
        entry = {}
        with self.assertRaises(KeyError):
            MachineModel("csx").set_instruction_entry(entry)
        with self.assertRaises(TypeError):
            MachineModel("csx").set_instruction()

    def test_sanity_check(self):
        output = StringIO()
        # non-verbose
        sanity_check("csx", verbose=False, internet_check=False, output_file=output)
        sanity_check("tx2", verbose=False, internet_check=False, output_file=output)
        sanity_check("zen1", verbose=False, internet_check=False, output_file=output)

        # verbose
        sanity_check("csx", verbose=True, internet_check=False, output_file=output)
        sanity_check("tx2", verbose=True, internet_check=False, output_file=output)
        sanity_check("zen1", verbose=True, internet_check=False, output_file=output)

    def test_ibench_import(self):
        # only check import without dumping the DB file (takes too much time)
        with open(self._find_file("ibench_import_x86.dat")) as input_file:
            input_data = input_file.readlines()
            entries = dbi._get_ibench_output(input_data, "x86")
            self.assertEqual(len(entries), 3)
            for _, e in entries.items():
                self.assertIsNotNone(e["throughput"])
                self.assertIsNotNone(e["latency"])
        with open(self._find_file("ibench_import_aarch64.dat")) as input_file:
            input_data = input_file.readlines()
            entries = dbi._get_ibench_output(input_data, "aarch64")
            self.assertEqual(len(entries), 4)
            for _, e in entries.items():
                self.assertIsNotNone(e["throughput"])
                self.assertIsNotNone(e["latency"])

    def test_asmbench_import(self):
        # only check import without dumping the DB file (takes too much time)
        with open(self._find_file("asmbench_import_x86.dat")) as input_file:
            input_data = input_file.readlines()
            entries = dbi._get_asmbench_output(input_data, "x86")
            self.assertEqual(len(entries), 3)
            for _, e in entries.items():
                self.assertIsNotNone(e["throughput"])
                self.assertIsNotNone(e["latency"])
        with open(self._find_file("asmbench_import_aarch64.dat")) as input_file:
            input_data = input_file.readlines()
            entries = dbi._get_asmbench_output(input_data, "aarch64")
            self.assertEqual(len(entries), 4)
            for _, e in entries.items():
                self.assertIsNotNone(e["throughput"])
                self.assertIsNotNone(e["latency"])
            # remove empty line => no import since broken format
            del input_data[3]
            entries = dbi._get_asmbench_output(input_data, "aarch64")
            self.assertEqual(len(entries), 0)
        with self.assertRaises(ValueError):
            dbi.import_benchmark_output(
                "csx", "invalid_bench_type", self._find_file("asmbench_import_x86.dat")
            )
        with self.assertRaises(AssertionError):
            dbi.import_benchmark_output("csx", "ibench", "invalid_file")

    def test_online_scraping(self):
        # addpd -- suspicious instruction, normal URL
        instr_1 = ["addpd", (True, "(r) (r,w)")]
        self.assertEqual(dbi._scrape_from_felixcloutier(instr_1[0]), instr_1[1])
        # movpd -- not suspicious,
        instr_2 = ["movapd", (False, "(r) (w)")]
        self.assertEqual(dbi._scrape_from_felixcloutier(instr_2[0]), instr_2[1])
        # vfmadd132pd -- only in combined view with 213/231.
        # No 2-operand version, therefore, empty string
        instr_3 = ["vfmadd132pd", (True, "")]
        self.assertEqual(dbi._scrape_from_felixcloutier(instr_3[0]), instr_3[1])

    ##################
    # Helper functions
    ##################
    @staticmethod
    def _find_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestDBInterface)
    unittest.TextTestRunner(verbosity=2, buffer=True).run(suite)



#!/usr/bin/env python3
from collections import OrderedDict

from osaca.parser import ParserAArch64, ParserX86ATT, get_parser

COMMENT_MARKER = {"start": "OSACA-BEGIN", "end": "OSACA-END"}


def reduce_to_section(kernel, isa):
    """
    Finds OSACA markers in given kernel and returns marked section

    :param list kernel: kernel to check
    :param str isa: ISA of given kernel
    :returns: `list` -- marked section of kernel as list of instruction forms
    """
    isa = isa.lower()
    if isa == "x86":
        start, end = find_marked_kernel_x86ATT(kernel)
    elif isa == "aarch64":
        start, end = find_marked_kernel_AArch64(kernel)
    else:
        raise ValueError("ISA not supported.")
    if start == -1:
        start = 0
    if end == -1:
        end = len(kernel)
    return kernel[start:end]


def find_marked_kernel_AArch64(lines):
    """
    Find marked section for AArch64

    :param list lines: kernel
    :returns: `tuple of int` -- start and end line of marked section
    """
    nop_bytes = ["213", "3", "32", "31"]
    return find_marked_section(
        lines,
        ParserAArch64(),
        ["mov"],
        "x1",
        [111, 222],
        nop_bytes,
        reverse=True,
        comments=COMMENT_MARKER,
    )


def find_marked_kernel_x86ATT(lines):
    """
    Find marked section for x86

    :param list lines: kernel
    :returns: `tuple of int` -- start and end line of marked section
    """
    nop_bytes = ["100", "103", "144"]
    return find_marked_section(
        lines,
        ParserX86ATT(),
        ["mov", "movl"],
        "ebx",
        [111, 222],
        nop_bytes,
        comments=COMMENT_MARKER,
    )


def get_marker(isa, comment=""):
    """Return tuple of start and end marker lines."""
    isa = isa.lower()
    if isa == "x86":
        start_marker_raw = (
            "movl      $111, %ebx # OSACA START MARKER\n"
            ".byte     100        # OSACA START MARKER\n"
            ".byte     103        # OSACA START MARKER\n"
            ".byte     144        # OSACA START MARKER\n"
        )
        if comment:
            start_marker_raw += "# {}\n".format(comment)
        end_marker_raw = (
            "movl      $222, %ebx # OSACA END MARKER\n"
            ".byte     100        # OSACA END MARKER\n"
            ".byte     103        # OSACA END MARKER\n"
            ".byte     144        # OSACA END MARKER\n"
        )
    elif isa == "aarch64":
        start_marker_raw = (
            "mov       x1, #111    // OSACA START MARKER\n"
            ".byte     213,3,32,31 // OSACA START MARKER\n"
        )
        if comment:
            start_marker_raw += "// {}\n".format(comment)
        # After loop
        end_marker_raw = (
            "mov       x1, #222    // OSACA END MARKER\n"
            ".byte     213,3,32,31 // OSACA END MARKER\n"
        )

    parser = get_parser(isa)
    start_marker = parser.parse_file(start_marker_raw)
    end_marker = parser.parse_file(end_marker_raw)

    return start_marker, end_marker


def find_marked_section(
    lines, parser, mov_instr, mov_reg, mov_vals, nop_bytes, reverse=False, comments=None
):
    """
    Return indexes of marked section

    :param list lines: kernel
    :param parser: parser to use for checking
    :type parser: :class:`~parser.BaseParser`
    :param mov_instr: all MOV instruction possible for the marker
    :type mov_instr: `list of str`
    :param mov_reg: register used for the marker
    :type mov_reg: `str`
    :param mov_vals: values needed to be moved to ``mov_reg`` for valid marker
    :type mov_vals: `list of int`
    :param nop_bytes: bytes representing opcode of NOP
    :type nop_bytes: `list of int`
    :param reverse: indicating if ISA syntax requires reverse operand order, defaults to `False`
    :type reverse: boolean, optional
    :param comments: dictionary with start and end markers in comment format, defaults to None
    :type comments: dict, optional
    :returns: `tuple of int` -- start and end line of marked section
    """
    # TODO match to instructions returned by get_marker
    index_start = -1
    index_end = -1
    for i, line in enumerate(lines):
        try:
            if line.instruction is None and comments is not None and line.comment is not None:
                if comments["start"] == line.comment:
                    index_start = i + 1
                elif comments["end"] == line.comment:
                    index_end = i
            elif line.instruction in mov_instr and lines[i + 1].directive is not None:
                source = line.operands[0 if not reverse else 1]
                destination = line.operands[1 if not reverse else 0]
                # instruction pair matches, check for operands
                if (
                    "immediate" in source
                    and parser.normalize_imd(source.immediate) == mov_vals[0]
                    and "register" in destination
                    and parser.get_full_reg_name(destination.register) == mov_reg
                ):
                    # operands of first instruction match start, check for second one
                    match, line_count = match_bytes(lines, i + 1, nop_bytes)
                    if match:
                        # return first line after the marker
                        index_start = i + 1 + line_count
                elif (
                    "immediate" in source
                    and parser.normalize_imd(source.immediate) == mov_vals[1]
                    and "register" in destination
                    and parser.get_full_reg_name(destination.register) == mov_reg
                ):
                    # operand of first instruction match end, check for second one
                    match, line_count = match_bytes(lines, i + 1, nop_bytes)
                    if match:
                        # return line of the marker
                        index_end = i
        except TypeError:
            print(i, line)
        if index_start != -1 and index_end != -1:
            break
    return index_start, index_end


def match_bytes(lines, index, byte_list):
    """Match bytes directives of markers"""
    # either all bytes are in one line or in separate ones
    extracted_bytes = []
    line_count = 0
    while (
        index < len(lines)
        and lines[index].directive is not None
        and lines[index].directive.name == "byte"
    ):
        line_count += 1
        extracted_bytes += lines[index].directive.parameters
        index += 1
    if extracted_bytes[0 : len(byte_list)] == byte_list:
        return True, line_count
    return False, -1


def find_jump_labels(lines):
    """
    Find and return all labels which are followed by instructions until the next label

    :return: OrderedDict of mapping from label name to associated line index
    """
    # 1. Identify labels and instructions until next label
    labels = OrderedDict()
    current_label = None
    for i, line in enumerate(lines):
        if line["label"] is not None:
            # When a new label is found, add to blocks dict
            labels[line["label"]] = (i,)
            # End previous block at previous line
            if current_label is not None:
                labels[current_label] = (labels[current_label][0], i)
            # Update current block name
            current_label = line["label"]
        elif current_label is None:
            # If no block has been started, skip end detection
            continue
    # Set to last line if no end was for last label found
    if current_label is not None and len(labels[current_label]) == 1:
        labels[current_label] = (labels[current_label][0], len(lines))

    # 2. Identify and remove labels which contain only dot-instructions (e.g., .text)
    for label in list(labels):
        if all(
            [
                line["instruction"].startswith(".")
                for line in lines[labels[label][0] : labels[label][1]]
                if line["instruction"] is not None
            ]
        ):
            del labels[label]

    return OrderedDict([(label, v[0]) for label, v in labels.items()])


def find_basic_blocks(lines):
    """
    Find and return basic blocks (asm sections which can only be executed as complete block).

    Blocks always start at a label and end at the next jump/break possibility.

    :return: OrderedDict with labels as keys and list of lines as value
    """
    valid_jump_labels = find_jump_labels(lines)

    # Identify blocks, as they are started with a valid jump label and terminated by a label or
    # an instruction referencing a valid jump label
    blocks = OrderedDict()
    for label, label_line_idx in valid_jump_labels.items():
        blocks[label] = [lines[label_line_idx]]
        for line in lines[label_line_idx + 1 :]:
            terminate = False
            blocks[label].append(line)
            # Find end of block by searching for references to valid jump labels
            if line["instruction"] and line["operands"]:
                for operand in [o for o in line["operands"] if "identifier" in o]:
                    if operand["identifier"]["name"] in valid_jump_labels:
                        terminate = True
            elif line["label"] is not None:
                terminate = True
            if terminate:
                break

    return blocks


def find_basic_loop_bodies(lines):
    """
    Find and return basic loop bodies (asm section which loop back on itself with no other egress).

    :return: OrderedDict with labels as keys and list of lines as value
    """
    valid_jump_labels = find_jump_labels(lines)

    # Identify blocks, as they are started with a valid jump label and terminated by
    # an instruction referencing a valid jump label
    loop_bodies = OrderedDict()
    for label, label_line_idx in valid_jump_labels.items():
        current_block = [lines[label_line_idx]]
        for line in lines[label_line_idx + 1 :]:
            terminate = False
            current_block.append(line)
            # Find end of block by searching for references to valid jump labels
            if line["instruction"] and line["operands"]:
                # Ignore `b.none` instructions (relevant von ARM SVE code)
                # This branch instruction is often present _within_ inner loop blocks, but usually
                # do not terminate
                if line["instruction"] == "b.none":
                    continue
                for operand in [o for o in line["operands"] if "identifier" in o]:
                    if operand["identifier"]["name"] in valid_jump_labels:
                        if operand["identifier"]["name"] == label:
                            loop_bodies[label] = current_block
                        terminate = True
                        break
            if terminate:
                break

    return loop_bodies



"""
Collection of parsers supported by OSACA.

Only the parser below will be exported, so please add new parsers to __all__.
"""
from .attr_dict import AttrDict
from .base_parser import BaseParser
from .parser_x86att import ParserX86ATT
from .parser_AArch64 import ParserAArch64

__all__ = ["AttrDict", "BaseParser", "ParserX86ATT", "ParserAArch64", "get_parser"]


def get_parser(isa):
    if isa.lower() == "x86":
        return ParserX86ATT()
    elif isa.lower() == "aarch64":
        return ParserAArch64()
    else:
        raise ValueError("Unknown ISA {!r}.".format(isa))



#!/usr/bin/env python3
"""
Unit tests for the CLI of OSACA and running the sample kernels in examples/
"""

import argparse
import os
import unittest
from io import StringIO
from shutil import copyfile
from unittest.mock import patch

import osaca.osaca as osaca
from osaca.db_interface import sanity_check
from osaca.parser import ParserAArch64, ParserX86ATT
from osaca.semantics import MachineModel


class ErrorRaisingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ValueError(message)  # reraise an error


class TestCLI(unittest.TestCase):

    ###########
    # Tests
    ###########

    def test_check_arguments(self):
        parser = osaca.create_parser(parser=ErrorRaisingArgumentParser())
        args = parser.parse_args(["--arch", "WRONG_ARCH", self._find_file("gs", "csx", "gcc")])
        with self.assertRaises(ValueError):
            osaca.check_arguments(args, parser)
        args = parser.parse_args(
            [
                "--arch",
                "csx",
                "--import",
                "WRONG_BENCH",
                self._find_file("gs", "csx", "gcc"),
            ]
        )
        with self.assertRaises(ValueError):
            osaca.check_arguments(args, parser)

    def test_import_data(self):
        parser = osaca.create_parser(parser=ErrorRaisingArgumentParser())
        args = parser.parse_args(
            [
                "--arch",
                "tx2",
                "--import",
                "ibench",
                self._find_test_file("ibench_import_aarch64.dat"),
            ]
        )
        output = StringIO()
        osaca.run(args, output_file=output)
        args = parser.parse_args(
            [
                "--arch",
                "tx2",
                "--import",
                "asmbench",
                self._find_test_file("asmbench_import_aarch64.dat"),
            ]
        )
        osaca.run(args, output_file=output)

    def test_check_db(self):
        parser = osaca.create_parser(parser=ErrorRaisingArgumentParser())
        args = parser.parse_args(
            [
                "--arch",
                "tx2",
                "--db-check",
                "--verbose",
                self._find_test_file("triad_x86_iaca.s"),
            ]
        )
        output = StringIO()
        osaca.run(args, output_file=output)

    def test_get_parser(self):
        self.assertTrue(isinstance(osaca.get_asm_parser("csx"), ParserX86ATT))
        self.assertTrue(isinstance(osaca.get_asm_parser("tx2"), ParserAArch64))
        with self.assertRaises(ValueError):
            osaca.get_asm_parser("UNKNOWN")

    def test_marker_insert_x86(self):
        # copy file to add markers
        name = self._find_test_file("kernel_x86.s")
        name_copy = name + ".copy.s"
        copyfile(name, name_copy)

        user_input = [".L10"]
        output = StringIO()
        parser = osaca.create_parser()
        args = parser.parse_args(["--arch", "csx", "--insert-marker", name_copy])
        with patch("builtins.input", side_effect=user_input):
            osaca.run(args, output_file=output)

        lines_orig = len(open(name).readlines())
        lines_copy = len(open(name_copy).readlines())
        self.assertEqual(lines_copy, lines_orig + 5 + 4)
        # remove copy again
        os.remove(name_copy)

    def test_marker_insert_aarch64(self):
        # copy file to add markers
        name = self._find_test_file("kernel_aarch64.s")
        name_copy = name + ".copy.s"
        copyfile(name, name_copy)

        user_input = [".LBB0_32", "64"]
        parser = osaca.create_parser()
        args = parser.parse_args(["--arch", "tx2", "--insert-marker", name_copy])
        with patch("builtins.input", side_effect=user_input):
            osaca.run(args)

        lines_orig = len(open(name).readlines())
        lines_copy = len(open(name_copy).readlines())
        self.assertEqual(lines_copy, lines_orig + 3 + 2)
        # remove copy again
        os.remove(name_copy)

    def test_examples(self):
        kernels = [
            "add",
            "copy",
            "daxpy",
            "gs",
            "j2d",
            "striad",
            "sum_reduction",
            "triad",
            "update",
        ]
        archs = ["csx", "tx2", "zen1"]
        comps = {"csx": ["gcc", "icc"], "tx2": ["gcc", "clang"], "zen1": ["gcc"]}
        parser = osaca.create_parser()
        # Analyze all asm files resulting out of kernels, archs and comps
        for k in kernels:
            for a in archs:
                for c in comps[a]:
                    with self.subTest(kernel=k, arch=a, comp=c):
                        args = parser.parse_args(
                            [
                                "--arch",
                                a,
                                self._find_file(k, a, c),
                                "--export-graph",
                                "/dev/null",
                            ]
                        )
                        output = StringIO()
                        osaca.run(args, output_file=output)
                        self.assertTrue("WARNING" not in output.getvalue())

    def test_architectures(self):
        parser = osaca.create_parser()
        # Run the test kernel for all architectures
        archs = osaca.SUPPORTED_ARCHS
        for arch in archs:
            with self.subTest(micro_arch=arch):
                isa = MachineModel.get_isa_for_arch(arch)
                kernel = "kernel_{}.s".format(isa)
                args = parser.parse_args(["--arch", arch, self._find_test_file(kernel)])
                output = StringIO()
                osaca.run(args, output_file=output)

    def test_architectures_sanity(self):
        # Run sanity check for all architectures
        archs = osaca.SUPPORTED_ARCHS
        for arch in archs:
            with self.subTest(micro_arch=arch):
                out = StringIO()
                sanity = sanity_check(arch, verbose=2, output_file=out)
                self.assertTrue(sanity, msg=out)

    def test_without_arch(self):
        # Run test kernels without --arch flag
        parser = osaca.create_parser()
        # x86
        kernel_x86 = "kernel_x86.s"
        args = parser.parse_args([self._find_test_file(kernel_x86)])
        output = StringIO()
        osaca.run(args, output_file=output)
        # AArch64
        kernel_aarch64 = "kernel_aarch64.s"
        args = parser.parse_args([self._find_test_file(kernel_aarch64)])
        osaca.run(args, output_file=output)

    def test_user_warnings(self):
        parser = osaca.create_parser()
        kernel = "triad_x86_unmarked.s"
        args = parser.parse_args(
            ["--arch", "csx", "--ignore-unknown", self._find_test_file(kernel)]
        )
        output = StringIO()
        osaca.run(args, output_file=output)
        # WARNING for length
        self.assertTrue(
            output.getvalue().count(
                "WARNING: You are analyzing a large amount of instruction forms"
            )
            == 1
        )
        # WARNING for arch
        args = parser.parse_args(
            ["--lines", "100-199", "--ignore-unknown", self._find_test_file(kernel)]
        )
        output = StringIO()
        osaca.run(args, output_file=output)
        self.assertTrue(
            output.getvalue().count("WARNING: No micro-architecture was specified") == 1
        )
        # WARNING for timeout
        args = parser.parse_args(
            ["--ignore-unknown", "--lcd-timeout", "0", self._find_test_file(kernel)]
        )
        output = StringIO()
        osaca.run(args, output_file=output)
        self.assertTrue(output.getvalue().count("WARNING: LCD analysis timed out") == 1)
        args = parser.parse_args(
            ["--ignore-unknown", "--lcd-timeout", "-1", self._find_test_file(kernel)]
        )
        output = StringIO()
        osaca.run(args, output_file=output)
        self.assertTrue(output.getvalue().count("WARNING: LCD analysis timed out") == 0)

    def test_lines_arg(self):
        # Run tests with --lines option
        parser = osaca.create_parser()
        kernel_x86 = "triad_x86_iaca.s"
        args_base = parser.parse_args(["--arch", "csx", self._find_test_file(kernel_x86)])
        output_base = StringIO()
        osaca.run(args_base, output_file=output_base)
        output_base = output_base.getvalue().split("\n")[8:]
        args = []
        args.append(
            parser.parse_args(
                [
                    "--lines",
                    "146-154",
                    "--arch",
                    "csx",
                    self._find_test_file(kernel_x86),
                ]
            )
        )
        args.append(
            parser.parse_args(
                [
                    "--lines",
                    "146:154",
                    "--arch",
                    "csx",
                    self._find_test_file(kernel_x86),
                ]
            )
        )
        args.append(
            parser.parse_args(
                [
                    "--lines",
                    "146,147:148,149-154",
                    "--arch",
                    "csx",
                    self._find_test_file(kernel_x86),
                ]
            )
        )
        for a in args:
            with self.subTest(params=a):
                output = StringIO()
                osaca.run(a, output_file=output)
                self.assertEqual(output.getvalue().split("\n")[8:], output_base)

    ##################
    # Helper functions
    ##################

    @staticmethod
    def _find_file(kernel, arch, comp):
        testdir = os.path.dirname(__file__)
        name = os.path.join(
            testdir,
            "../examples",
            kernel,
            kernel + ".s." + arch[:3].lower() + "." + comp.lower() + ".s",
        )
        if kernel == "j2d" and arch.lower() == "csx":
            name = name[:-1] + "AVX.s"
        assert os.path.exists(name)
        return name

    @staticmethod
    def _find_test_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestCLI)
    unittest.TextTestRunner(verbosity=2, buffer=True).run(suite)



#!/usr/bin/env python3
"""
Unit tests for Semantic Analysis
"""

import os
import unittest
import time
from copy import deepcopy

import networkx as nx
from osaca.osaca import get_unmatched_instruction_ratio
from osaca.parser import AttrDict, ParserAArch64, ParserX86ATT
from osaca.semantics import (
    INSTR_FLAGS,
    ArchSemantics,
    ISASemantics,
    KernelDG,
    MachineModel,
    reduce_to_section,
)


class TestSemanticTools(unittest.TestCase):
    MODULE_DATA_DIR = os.path.join(
        os.path.dirname(os.path.split(os.path.abspath(__file__))[0]), "osaca/data/"
    )

    @classmethod
    def setUpClass(cls):
        # set up parser and kernels
        cls.parser_x86 = ParserX86ATT()
        cls.parser_AArch64 = ParserAArch64()
        with open(cls._find_file("kernel_x86.s")) as f:
            cls.code_x86 = f.read()
        with open(cls._find_file("kernel_x86_memdep.s")) as f:
            cls.code_x86_memdep = f.read()
        with open(cls._find_file("kernel_x86_long_LCD.s")) as f:
            cls.code_x86_long_LCD = f.read()
        with open(cls._find_file("kernel_aarch64_memdep.s")) as f:
            cls.code_aarch64_memdep = f.read()
        with open(cls._find_file("kernel_aarch64.s")) as f:
            cls.code_AArch64 = f.read()
        with open(cls._find_file("kernel_aarch64_sve.s")) as f:
            cls.code_AArch64_SVE = f.read()
        cls.kernel_x86 = reduce_to_section(cls.parser_x86.parse_file(cls.code_x86), "x86")
        cls.kernel_x86_memdep = reduce_to_section(
            cls.parser_x86.parse_file(cls.code_x86_memdep), "x86"
        )
        cls.kernel_x86_long_LCD = reduce_to_section(
            cls.parser_x86.parse_file(cls.code_x86_long_LCD), "x86"
        )
        cls.kernel_AArch64 = reduce_to_section(
            cls.parser_AArch64.parse_file(cls.code_AArch64), "aarch64"
        )
        cls.kernel_aarch64_memdep = reduce_to_section(
            cls.parser_AArch64.parse_file(cls.code_aarch64_memdep), "aarch64"
        )
        cls.kernel_aarch64_SVE = reduce_to_section(
            cls.parser_AArch64.parse_file(cls.code_AArch64_SVE), "aarch64"
        )

        # set up machine models
        cls.machine_model_csx = MachineModel(
            path_to_yaml=os.path.join(cls.MODULE_DATA_DIR, "csx.yml")
        )
        cls.machine_model_tx2 = MachineModel(
            path_to_yaml=os.path.join(cls.MODULE_DATA_DIR, "tx2.yml")
        )
        cls.machine_model_a64fx = MachineModel(
            path_to_yaml=os.path.join(cls.MODULE_DATA_DIR, "a64fx.yml")
        )
        cls.semantics_x86 = ISASemantics("x86")
        cls.semantics_csx = ArchSemantics(
            cls.machine_model_csx,
            path_to_yaml=os.path.join(cls.MODULE_DATA_DIR, "isa/x86.yml"),
        )
        cls.semantics_aarch64 = ISASemantics("aarch64")
        cls.semantics_tx2 = ArchSemantics(
            cls.machine_model_tx2,
            path_to_yaml=os.path.join(cls.MODULE_DATA_DIR, "isa/aarch64.yml"),
        )
        cls.semantics_a64fx = ArchSemantics(
            cls.machine_model_a64fx,
            path_to_yaml=os.path.join(cls.MODULE_DATA_DIR, "isa/aarch64.yml"),
        )
        cls.machine_model_zen = MachineModel(arch="zen1")

        for i in range(len(cls.kernel_x86)):
            cls.semantics_csx.assign_src_dst(cls.kernel_x86[i])
            cls.semantics_csx.assign_tp_lt(cls.kernel_x86[i])
        for i in range(len(cls.kernel_x86_memdep)):
            cls.semantics_csx.assign_src_dst(cls.kernel_x86_memdep[i])
            cls.semantics_csx.assign_tp_lt(cls.kernel_x86_memdep[i])
        for i in range(len(cls.kernel_x86_long_LCD)):
            cls.semantics_csx.assign_src_dst(cls.kernel_x86_long_LCD[i])
            cls.semantics_csx.assign_tp_lt(cls.kernel_x86_long_LCD[i])
        for i in range(len(cls.kernel_AArch64)):
            cls.semantics_tx2.assign_src_dst(cls.kernel_AArch64[i])
            cls.semantics_tx2.assign_tp_lt(cls.kernel_AArch64[i])
        for i in range(len(cls.kernel_aarch64_memdep)):
            cls.semantics_tx2.assign_src_dst(cls.kernel_aarch64_memdep[i])
            cls.semantics_tx2.assign_tp_lt(cls.kernel_aarch64_memdep[i])
        for i in range(len(cls.kernel_aarch64_SVE)):
            cls.semantics_a64fx.assign_src_dst(cls.kernel_aarch64_SVE[i])
            cls.semantics_a64fx.assign_tp_lt(cls.kernel_aarch64_SVE[i])

    ###########
    # Tests
    ###########

    def test_creation_by_name(self):
        try:
            tmp_mm = MachineModel(arch="CSX")
            ArchSemantics(tmp_mm)
        except ValueError:
            self.fail()

    def test_machine_model_various_functions(self):
        # check dummy MachineModel creation
        try:
            MachineModel(isa="x86")
            MachineModel(isa="aarch64")
        except ValueError:
            self.fail()
        test_mm_x86 = MachineModel(path_to_yaml=self._find_file("test_db_x86.yml"))
        test_mm_arm = MachineModel(path_to_yaml=self._find_file("test_db_aarch64.yml"))

        # test get_instruction without mnemonic
        self.assertIsNone(test_mm_x86.get_instruction(None, []))
        self.assertIsNone(test_mm_arm.get_instruction(None, []))

        # test get_instruction from DB
        self.assertIsNone(test_mm_x86.get_instruction(None, []))
        self.assertIsNone(test_mm_arm.get_instruction(None, []))
        self.assertIsNone(test_mm_x86.get_instruction("NOT_IN_DB", []))
        self.assertIsNone(test_mm_arm.get_instruction("NOT_IN_DB", []))
        name_x86_1 = "vaddpd"
        operands_x86_1 = [
            {"class": "register", "name": "xmm"},
            {"class": "register", "name": "xmm"},
            {"class": "register", "name": "xmm"},
        ]
        instr_form_x86_1 = test_mm_x86.get_instruction(name_x86_1, operands_x86_1)
        self.assertEqual(instr_form_x86_1, test_mm_x86.get_instruction(name_x86_1, operands_x86_1))
        self.assertEqual(
            test_mm_x86.get_instruction("jg", [{"class": "identifier"}]),
            test_mm_x86.get_instruction("jg", [{"class": "identifier"}]),
        )
        name_arm_1 = "fadd"
        operands_arm_1 = [
            {"class": "register", "prefix": "v", "shape": "s"},
            {"class": "register", "prefix": "v", "shape": "s"},
            {"class": "register", "prefix": "v", "shape": "s"},
        ]
        instr_form_arm_1 = test_mm_arm.get_instruction(name_arm_1, operands_arm_1)
        self.assertEqual(instr_form_arm_1, test_mm_arm.get_instruction(name_arm_1, operands_arm_1))
        self.assertEqual(
            test_mm_arm.get_instruction("b.ne", [{"class": "identifier"}]),
            test_mm_arm.get_instruction("b.ne", [{"class": "identifier"}]),
        )

        # test full instruction name
        self.assertEqual(
            MachineModel.get_full_instruction_name(instr_form_x86_1),
            "vaddpd  register(name:xmm),register(name:xmm),register(name:xmm)",
        )
        self.assertEqual(
            MachineModel.get_full_instruction_name(instr_form_arm_1),
            "fadd  register(prefix:v,shape:s),register(prefix:v,shape:s),"
            + "register(prefix:v,shape:s)",
        )

        # test get_store_tp
        self.assertEqual(
            test_mm_x86.get_store_throughput(
                {"base": {"name": "x"}, "offset": None, "index": None, "scale": 1}
            ),
            [[2, "237"], [2, "4"]],
        )
        self.assertEqual(
            test_mm_x86.get_store_throughput(
                {
                    "base": {"prefix": "NOT_IN_DB"},
                    "offset": None,
                    "index": "NOT_NONE",
                    "scale": 1,
                }
            ),
            [[1, "23"], [1, "4"]],
        )
        self.assertEqual(
            test_mm_arm.get_store_throughput(
                {"base": {"prefix": "x"}, "offset": None, "index": None, "scale": 1}
            ),
            [[2, "34"], [2, "5"]],
        )
        self.assertEqual(
            test_mm_arm.get_store_throughput(
                {
                    "base": {"prefix": "NOT_IN_DB"},
                    "offset": None,
                    "index": None,
                    "scale": 1,
                }
            ),
            [[1, "34"], [1, "5"]],
        )

        # test get_store_lt
        self.assertEqual(
            test_mm_x86.get_store_latency(
                {"base": {"name": "x"}, "offset": None, "index": None, "scale": "1"}
            ),
            0,
        )
        self.assertEqual(
            test_mm_arm.get_store_latency(
                {"base": {"prefix": "x"}, "offset": None, "index": None, "scale": "1"}
            ),
            0,
        )

        # test has_hidden_load
        self.assertFalse(test_mm_x86.has_hidden_loads())

        # test default load tp
        self.assertEqual(
            test_mm_x86.get_load_throughput(
                {"base": {"name": "x"}, "offset": None, "index": None, "scale": 1}
            ),
            [[1, "23"], [1, ["2D", "3D"]]],
        )

        # test adding port
        test_mm_x86.add_port("dummyPort")
        test_mm_arm.add_port("dummyPort")

        # test dump of DB
        with open("/dev/null", "w") as dev_null:
            test_mm_x86.dump(stream=dev_null)
            test_mm_arm.dump(stream=dev_null)

    def test_src_dst_assignment_x86(self):
        for instruction_form in self.kernel_x86:
            with self.subTest(instruction_form=instruction_form):
                if instruction_form["semantic_operands"] is not None:
                    self.assertTrue("source" in instruction_form["semantic_operands"])
                    self.assertTrue("destination" in instruction_form["semantic_operands"])
                    self.assertTrue("src_dst" in instruction_form["semantic_operands"])

    def test_src_dst_assignment_AArch64(self):
        for instruction_form in self.kernel_AArch64:
            with self.subTest(instruction_form=instruction_form):
                if instruction_form["semantic_operands"] is not None:
                    self.assertTrue("source" in instruction_form["semantic_operands"])
                    self.assertTrue("destination" in instruction_form["semantic_operands"])
                    self.assertTrue("src_dst" in instruction_form["semantic_operands"])

    def test_tp_lt_assignment_x86(self):
        self.assertTrue("ports" in self.machine_model_csx)
        port_num = len(self.machine_model_csx["ports"])
        for instruction_form in self.kernel_x86:
            with self.subTest(instruction_form=instruction_form):
                self.assertTrue("throughput" in instruction_form)
                self.assertTrue("latency" in instruction_form)
                self.assertIsInstance(instruction_form["port_pressure"], list)
                self.assertEqual(len(instruction_form["port_pressure"]), port_num)

    def test_tp_lt_assignment_AArch64(self):
        self.assertTrue("ports" in self.machine_model_tx2)
        port_num = len(self.machine_model_tx2["ports"])
        for instruction_form in self.kernel_AArch64:
            with self.subTest(instruction_form=instruction_form):
                self.assertTrue("throughput" in instruction_form)
                self.assertTrue("latency" in instruction_form)
                self.assertIsInstance(instruction_form["port_pressure"], list)
                self.assertEqual(len(instruction_form["port_pressure"]), port_num)

    def test_optimal_throughput_assignment(self):
        # x86
        kernel_fixed = deepcopy(self.kernel_x86)
        self.semantics_csx.add_semantics(kernel_fixed)
        self.assertEqual(get_unmatched_instruction_ratio(kernel_fixed), 0)
        kernel_optimal = deepcopy(kernel_fixed)
        self.semantics_csx.assign_optimal_throughput(kernel_optimal)
        tp_fixed = self.semantics_csx.get_throughput_sum(kernel_fixed)
        tp_optimal = self.semantics_csx.get_throughput_sum(kernel_optimal)
        self.assertNotEqual(tp_fixed, tp_optimal)
        self.assertTrue(max(tp_optimal) <= max(tp_fixed))

        # arm
        kernel_fixed = deepcopy(self.kernel_AArch64)
        self.semantics_tx2.add_semantics(kernel_fixed)
        self.assertEqual(get_unmatched_instruction_ratio(kernel_fixed), 0)
        kernel_optimal = deepcopy(kernel_fixed)
        self.semantics_tx2.assign_optimal_throughput(kernel_optimal)
        tp_fixed = self.semantics_tx2.get_throughput_sum(kernel_fixed)
        tp_optimal = self.semantics_tx2.get_throughput_sum(kernel_optimal)
        self.assertNotEqual(tp_fixed, tp_optimal)
        self.assertTrue(max(tp_optimal) <= max(tp_fixed))

    def test_kernelDG_x86(self):
        #
        #  4
        #   \___>6__>7
        #   /
        #  3
        #     5_______>9
        #
        dg = KernelDG(self.kernel_x86, self.parser_x86, self.machine_model_csx, self.semantics_csx)
        self.assertTrue(nx.algorithms.dag.is_directed_acyclic_graph(dg.dg))
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=3))), 1)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=3)), 6)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=4))), 1)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=4)), 6)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=5))), 1)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=5)), 9)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=6))), 1)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=6)), 7)
        self.assertEqual(list(dg.get_dependent_instruction_forms(line_number=7)), [])
        self.assertEqual(list(dg.get_dependent_instruction_forms(line_number=8)), [])
        with self.assertRaises(ValueError):
            dg.get_dependent_instruction_forms()
        # test dot creation
        dg.export_graph(filepath="/dev/null")

    def test_memdependency_x86(self):
        dg = KernelDG(
            self.kernel_x86_memdep,
            self.parser_x86,
            self.machine_model_csx,
            self.semantics_csx,
        )
        self.assertTrue(nx.algorithms.dag.is_directed_acyclic_graph(dg.dg))
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=3)), {6, 8})
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=5)), {10, 12})
        with self.assertRaises(ValueError):
            dg.get_dependent_instruction_forms()
        # test dot creation
        dg.export_graph(filepath="/dev/null")

    def test_kernelDG_AArch64(self):
        dg = KernelDG(
            self.kernel_AArch64,
            self.parser_AArch64,
            self.machine_model_tx2,
            self.semantics_tx2,
        )
        self.assertTrue(nx.algorithms.dag.is_directed_acyclic_graph(dg.dg))
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=3)), {7, 8})
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=4)), {9, 10})
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=5)), {7, 8})
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=6)), {9, 10})
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=7)), 13)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=8)), 14)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=9)), 16)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=10)), 17)
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=11)), {13, 14})
        self.assertEqual(set(dg.get_dependent_instruction_forms(line_number=12)), {16, 17})
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=13)), 15)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=14)), 15)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=15))), 0)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=16)), 18)
        self.assertEqual(next(dg.get_dependent_instruction_forms(line_number=17)), 18)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=18))), 0)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=19))), 0)
        self.assertEqual(len(list(dg.get_dependent_instruction_forms(line_number=20))), 0)
        with self.assertRaises(ValueError):
            dg.get_dependent_instruction_forms()
        # test dot creation
        dg.export_graph(filepath="/dev/null")

    def test_kernelDG_SVE(self):
        KernelDG(
            self.kernel_aarch64_SVE,
            self.parser_AArch64,
            self.machine_model_a64fx,
            self.semantics_a64fx,
        )
        # TODO check for correct analysis

    def test_hidden_load(self):
        machine_model_hld = MachineModel(
            path_to_yaml=self._find_file("hidden_load_machine_model.yml")
        )
        self.assertTrue(machine_model_hld.has_hidden_loads())
        semantics_hld = ArchSemantics(machine_model_hld)
        kernel_hld = self.parser_x86.parse_file(self.code_x86)
        kernel_hld_2 = self.parser_x86.parse_file(self.code_x86)
        kernel_hld_2 = self.parser_x86.parse_file(self.code_x86)[-3:]
        kernel_hld_3 = self.parser_x86.parse_file(self.code_x86)[5:8]
        semantics_hld.add_semantics(kernel_hld)
        semantics_hld.add_semantics(kernel_hld_2)
        semantics_hld.add_semantics(kernel_hld_3)

        num_hidden_loads = len([x for x in kernel_hld if INSTR_FLAGS.HIDDEN_LD in x["flags"]])
        num_hidden_loads_2 = len([x for x in kernel_hld_2 if INSTR_FLAGS.HIDDEN_LD in x["flags"]])
        num_hidden_loads_3 = len([x for x in kernel_hld_3 if INSTR_FLAGS.HIDDEN_LD in x["flags"]])
        self.assertEqual(num_hidden_loads, 1)
        self.assertEqual(num_hidden_loads_2, 0)
        self.assertEqual(num_hidden_loads_3, 1)

    def test_cyclic_dag(self):
        dg = KernelDG(self.kernel_x86, self.parser_x86, self.machine_model_csx, self.semantics_csx)
        dg.dg.add_edge(100, 101, latency=1.0)
        dg.dg.add_edge(101, 102, latency=2.0)
        dg.dg.add_edge(102, 100, latency=3.0)
        with self.assertRaises(NotImplementedError):
            dg.get_critical_path()
        with self.assertRaises(NotImplementedError):
            dg.get_loopcarried_dependencies()

    def test_loop_carried_dependency_aarch64(self):
        dg = KernelDG(
            self.kernel_aarch64_memdep,
            self.parser_AArch64,
            self.machine_model_tx2,
            self.semantics_tx2,
        )
        lc_deps = dg.get_loopcarried_dependencies()
        self.assertEqual(len(lc_deps), 2)
        # based on line 6
        self.assertEqual(lc_deps[6]["latency"], 28.0)
        self.assertEqual(
            [(iform.line_number, lat) for iform, lat in lc_deps[6]["dependencies"]],
            [(6, 4.0), (10, 6.0), (11, 6.0), (12, 6.0), (13, 6.0), (14, 0)],
        )

    def test_loop_carried_dependency_x86(self):
        lcd_id = 8
        lcd_id2 = 5
        dg = KernelDG(self.kernel_x86, self.parser_x86, self.machine_model_csx, self.semantics_csx)
        lc_deps = dg.get_loopcarried_dependencies()
        self.assertEqual(len(lc_deps), 2)
        # ID 8
        self.assertEqual(
            lc_deps[lcd_id]["root"], dg.dg.nodes(data=True)[lcd_id]["instruction_form"]
        )
        self.assertEqual(len(lc_deps[lcd_id]["dependencies"]), 1)
        self.assertEqual(
            lc_deps[lcd_id]["dependencies"][0][0],
            dg.dg.nodes(data=True)[lcd_id]["instruction_form"],
        )
        # w/  flag dependencies: ID 9 w/ len=2
        # w/o flag dependencies: ID 5 w/ len=1
        # TODO discuss
        self.assertEqual(
            lc_deps[lcd_id2]["root"],
            dg.dg.nodes(data=True)[lcd_id2]["instruction_form"],
        )
        self.assertEqual(len(lc_deps[lcd_id2]["dependencies"]), 1)
        self.assertEqual(
            lc_deps[lcd_id2]["dependencies"][0][0],
            dg.dg.nodes(data=True)[lcd_id2]["instruction_form"],
        )

    def test_timeout_during_loop_carried_dependency(self):
        start_time = time.perf_counter()
        KernelDG(
            self.kernel_x86_long_LCD,
            self.parser_x86,
            self.machine_model_csx,
            self.semantics_x86,
            timeout=10,
        )
        end_time = time.perf_counter()
        time_10 = end_time - start_time
        start_time = time.perf_counter()
        KernelDG(
            self.kernel_x86_long_LCD,
            self.parser_x86,
            self.machine_model_csx,
            self.semantics_x86,
            timeout=2,
        )
        end_time = time.perf_counter()
        time_2 = end_time - start_time
        self.assertTrue(time_10 > 10)
        self.assertTrue(2 < time_2)
        self.assertTrue(time_2 < (time_10 - 7))

    def test_is_read_is_written_x86(self):
        # independent form HW model
        dag = KernelDG(self.kernel_x86, self.parser_x86, None, None)
        reg_rcx = AttrDict({"name": "rcx"})
        reg_ymm1 = AttrDict({"name": "ymm1"})

        instr_form_r_c = self.parser_x86.parse_line("vmovsd  %xmm0, (%r15,%rcx,8)")
        self.semantics_csx.assign_src_dst(instr_form_r_c)
        instr_form_non_r_c = self.parser_x86.parse_line("movl  %xmm0, (%r15,%rax,8)")
        self.semantics_csx.assign_src_dst(instr_form_non_r_c)
        instr_form_w_c = self.parser_x86.parse_line("movi $0x05ACA, %rcx")
        self.semantics_csx.assign_src_dst(instr_form_w_c)

        instr_form_rw_ymm_1 = self.parser_x86.parse_line("vinsertf128 $0x1, %xmm1, %ymm0, %ymm1")
        self.semantics_csx.assign_src_dst(instr_form_rw_ymm_1)
        instr_form_rw_ymm_2 = self.parser_x86.parse_line("vinsertf128 $0x1, %xmm0, %ymm1, %ymm1")
        self.semantics_csx.assign_src_dst(instr_form_rw_ymm_2)
        instr_form_r_ymm = self.parser_x86.parse_line("vmovapd %ymm1, %ymm0")
        self.semantics_csx.assign_src_dst(instr_form_r_ymm)

        self.assertTrue(dag.is_read(reg_rcx, instr_form_r_c))
        self.assertFalse(dag.is_read(reg_rcx, instr_form_non_r_c))
        self.assertFalse(dag.is_read(reg_rcx, instr_form_w_c))
        self.assertTrue(dag.is_written(reg_rcx, instr_form_w_c))
        self.assertFalse(dag.is_written(reg_rcx, instr_form_r_c))

        self.assertTrue(dag.is_read(reg_ymm1, instr_form_rw_ymm_1))
        self.assertTrue(dag.is_read(reg_ymm1, instr_form_rw_ymm_2))
        self.assertTrue(dag.is_read(reg_ymm1, instr_form_r_ymm))
        self.assertTrue(dag.is_written(reg_ymm1, instr_form_rw_ymm_1))
        self.assertTrue(dag.is_written(reg_ymm1, instr_form_rw_ymm_2))
        self.assertFalse(dag.is_written(reg_ymm1, instr_form_r_ymm))

    def test_is_read_is_written_AArch64(self):
        # independent form HW model
        dag = KernelDG(self.kernel_AArch64, self.parser_AArch64, None, None)
        reg_x1 = AttrDict({"prefix": "x", "name": "1"})
        reg_w1 = AttrDict({"prefix": "w", "name": "1"})
        reg_d1 = AttrDict({"prefix": "d", "name": "1"})
        reg_q1 = AttrDict({"prefix": "q", "name": "1"})
        reg_v1 = AttrDict({"prefix": "v", "name": "1", "lanes": "2", "shape": "d"})
        regs = [reg_d1, reg_q1, reg_v1]
        regs_gp = [reg_w1, reg_x1]

        instr_form_r_1 = self.parser_AArch64.parse_line("stp q1, q3, [x12, #192]")
        self.semantics_tx2.assign_src_dst(instr_form_r_1)
        instr_form_r_2 = self.parser_AArch64.parse_line("fadd v2.2d, v1.2d, v0.2d")
        self.semantics_tx2.assign_src_dst(instr_form_r_2)
        instr_form_w_1 = self.parser_AArch64.parse_line("ldr d1, [x1, #:got_lo12:q2c]")
        self.semantics_tx2.assign_src_dst(instr_form_w_1)
        instr_form_non_w_1 = self.parser_AArch64.parse_line("ldr x1, [x1, #:got_lo12:q2c]")
        self.semantics_tx2.assign_src_dst(instr_form_non_w_1)
        instr_form_rw_1 = self.parser_AArch64.parse_line("fmul v1.2d, v1.2d, v0.2d")
        self.semantics_tx2.assign_src_dst(instr_form_rw_1)
        instr_form_rw_2 = self.parser_AArch64.parse_line("ldp q2, q4, [x1, #64]!")
        self.semantics_tx2.assign_src_dst(instr_form_rw_2)
        instr_form_rw_3 = self.parser_AArch64.parse_line("str x4, [x1], #64")
        self.semantics_tx2.assign_src_dst(instr_form_rw_3)
        instr_form_non_rw_1 = self.parser_AArch64.parse_line("adds x1, x11")
        self.semantics_tx2.assign_src_dst(instr_form_non_rw_1)

        for reg in regs:
            with self.subTest(reg=reg):
                self.assertTrue(dag.is_read(reg, instr_form_r_1))
                self.assertTrue(dag.is_read(reg, instr_form_r_2))
                self.assertTrue(dag.is_read(reg, instr_form_rw_1))
                self.assertFalse(dag.is_read(reg, instr_form_rw_2))
                self.assertFalse(dag.is_read(reg, instr_form_rw_3))
                self.assertFalse(dag.is_read(reg, instr_form_w_1))
                self.assertTrue(dag.is_written(reg, instr_form_w_1))
                self.assertTrue(dag.is_written(reg, instr_form_rw_1))
                self.assertFalse(dag.is_written(reg, instr_form_non_w_1))
                self.assertFalse(dag.is_written(reg, instr_form_rw_2))
                self.assertFalse(dag.is_written(reg, instr_form_rw_3))
                self.assertFalse(dag.is_written(reg, instr_form_non_rw_1))
                self.assertFalse(dag.is_written(reg, instr_form_non_rw_1))
        for reg in regs_gp:
            with self.subTest(reg=reg):
                self.assertFalse(dag.is_read(reg, instr_form_r_1))
                self.assertFalse(dag.is_read(reg, instr_form_r_2))
                self.assertFalse(dag.is_read(reg, instr_form_rw_1))
                self.assertTrue(dag.is_read(reg, instr_form_rw_2))
                self.assertTrue(dag.is_read(reg, instr_form_rw_3))
                self.assertTrue(dag.is_read(reg, instr_form_w_1))
                self.assertFalse(dag.is_written(reg, instr_form_w_1))
                self.assertFalse(dag.is_written(reg, instr_form_rw_1))
                self.assertTrue(dag.is_written(reg, instr_form_non_w_1))
                self.assertTrue(dag.is_written(reg, instr_form_rw_2))
                self.assertTrue(dag.is_written(reg, instr_form_rw_3))
                self.assertTrue(dag.is_written(reg, instr_form_non_rw_1))
                self.assertTrue(dag.is_written(reg, instr_form_non_rw_1))

    def test_invalid_MachineModel(self):
        with self.assertRaises(ValueError):
            MachineModel()
        with self.assertRaises(ValueError):
            MachineModel(arch="CSX", path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "csx.yml"))
        with self.assertRaises(FileNotFoundError):
            MachineModel(arch="THE_MACHINE")
        with self.assertRaises(FileNotFoundError):
            MachineModel(path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "THE_MACHINE.yml"))

    def test_MachineModel_getter(self):
        sample_operands = [
            {
                "memory": {
                    "offset": None,
                    "base": {"name": "r12"},
                    "index": {"name": "rcx"},
                    "scale": 8,
                }
            }
        ]
        self.assertIsNone(self.machine_model_csx.get_instruction("GETRESULT", sample_operands))
        self.assertIsNone(self.machine_model_tx2.get_instruction("GETRESULT", sample_operands))

        self.assertEqual(self.machine_model_csx.get_arch(), "csx")
        self.assertEqual(self.machine_model_tx2.get_arch(), "tx2")

        self.assertEqual(self.machine_model_csx.get_ISA(), "x86")
        self.assertEqual(self.machine_model_tx2.get_ISA(), "aarch64")

        ports_csx = ["0", "0DV", "1", "2", "2D", "3", "3D", "4", "5", "6", "7"]
        data_ports_csx = ["2D", "3D"]
        self.assertEqual(self.machine_model_csx.get_ports(), ports_csx)
        self.assertEqual(self.machine_model_csx.get_data_ports(), data_ports_csx)

        self.assertFalse(self.machine_model_tx2.has_hidden_loads())

        self.assertEqual(MachineModel.get_isa_for_arch("CSX"), "x86")
        self.assertEqual(MachineModel.get_isa_for_arch("tX2"), "aarch64")
        with self.assertRaises(ValueError):
            self.assertIsNone(MachineModel.get_isa_for_arch("THE_MACHINE"))

    ##################
    # Helper functions
    ##################

    @staticmethod
    def _find_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSemanticTools)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python3
"""
Unit tests for IACA/OSACA marker utilities
"""
import os
import unittest
from collections import OrderedDict

from osaca.semantics import (
    reduce_to_section,
    find_basic_blocks,
    find_jump_labels,
    find_basic_loop_bodies,
)
from osaca.parser import ParserAArch64, ParserX86ATT


class TestMarkerUtils(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.parser_AArch = ParserAArch64()
        self.parser_x86 = ParserX86ATT()
        with open(self._find_file("triad_arm_iaca.s")) as f:
            triad_code_arm = f.read()
        with open(self._find_file("triad_x86_iaca.s")) as f:
            triad_code_x86 = f.read()
        self.parsed_AArch = self.parser_AArch.parse_file(triad_code_arm)
        self.parsed_x86 = self.parser_x86.parse_file(triad_code_x86)

    #################
    # Test
    #################

    def test_marker_detection_AArch64(self):
        kernel = reduce_to_section(self.parsed_AArch, "AArch64")
        self.assertEqual(len(kernel), 138)
        self.assertEqual(kernel[0].line_number, 307)
        self.assertEqual(kernel[-1].line_number, 444)

    def test_marker_detection_x86(self):
        kernel = reduce_to_section(self.parsed_x86, "x86")
        self.assertEqual(len(kernel), 9)
        self.assertEqual(kernel[0].line_number, 146)
        self.assertEqual(kernel[-1].line_number, 154)

    def test_marker_matching_AArch64(self):
        # preparation
        bytes_1_line = ".byte     213,3,32,31\n"
        bytes_2_lines_1 = ".byte     213,3,32\n" + ".byte 31\n"
        bytes_2_lines_2 = ".byte     213,3\n" + ".byte 32,31\n"
        bytes_2_lines_3 = ".byte     213\n" + ".byte 3,32,31\n"
        bytes_3_lines_1 = ".byte     213,3\n" + ".byte     32\n" + ".byte     31\n"
        bytes_3_lines_2 = ".byte     213\n" + ".byte     3,32\n" + ".byte     31\n"
        bytes_3_lines_3 = ".byte     213\n" + ".byte     3\n" + ".byte     32,31\n"
        bytes_4_lines = ".byte     213\n" + ".byte     3\n" + ".byte     32\n" + ".byte     31\n"
        mov_start_1 = "mov      x1, #111\n"
        mov_start_2 = "mov      x1, 111  // should work as well\n"
        mov_end_1 = "mov      x1, #222 // preferred way\n"
        mov_end_2 = "mov      x1, 222\n"
        prologue = (
            "mov x12, xzr\n"
            + "\tldp x9, x10, [sp, #16]      // 8-byte Folded Reload\n"
            + "     .p2align    6\n"
        )
        kernel = (
            ".LBB0_28:\n"
            + "fmul    v7.2d, v7.2d, v19.2d\n"
            + "stp q0, q1, [x10, #-32]\n"
            + "b.ne    .LBB0_28\n"
        )
        epilogue = ".LBB0_29:   //   Parent Loop BB0_20 Depth=1\n" + "bl    dummy\n"
        kernel_length = len(list(filter(None, kernel.split("\n"))))

        bytes_variations = [
            bytes_1_line,
            bytes_2_lines_1,
            bytes_2_lines_2,
            bytes_2_lines_3,
            bytes_3_lines_1,
            bytes_3_lines_2,
            bytes_3_lines_3,
            bytes_4_lines,
        ]
        mov_start_variations = [mov_start_1, mov_start_2]
        mov_end_variations = [mov_end_1, mov_end_2]
        # actual tests
        for mov_start_var in mov_start_variations:
            for bytes_var_1 in bytes_variations:
                for mov_end_var in mov_end_variations:
                    for bytes_var_2 in bytes_variations:
                        sample_code = (
                            prologue
                            + mov_start_var
                            + bytes_var_1
                            + kernel
                            + mov_end_var
                            + bytes_var_2
                            + epilogue
                        )
                        with self.subTest(
                            mov_start=mov_start_var,
                            bytes_start=bytes_var_1,
                            mov_end=mov_end_var,
                            bytes_end=bytes_var_2,
                        ):
                            sample_parsed = self.parser_AArch.parse_file(sample_code)
                            sample_kernel = reduce_to_section(sample_parsed, "AArch64")
                            self.assertEqual(len(sample_kernel), kernel_length)
                            kernel_start = len(
                                list(
                                    filter(
                                        None,
                                        (prologue + mov_start_var + bytes_var_1).split("\n"),
                                    )
                                )
                            )
                            parsed_kernel = self.parser_AArch.parse_file(
                                kernel, start_line=kernel_start
                            )
                            self.assertEqual(sample_kernel, parsed_kernel)

    def test_marker_matching_x86(self):
        # preparation
        bytes_1_line = ".byte     100,103,144\n"
        bytes_2_lines_1 = ".byte     100,103\n" + ".byte 144\n"
        bytes_2_lines_2 = ".byte     100\n" + ".byte 103,144\n"
        bytes_3_lines = (
            ".byte     100 # IACA MARKER UTILITY\n"
            + ".byte     103 # IACA MARKER UTILITY\n"
            + ".byte     144 # IACA MARKER UTILITY\n"
        )
        mov_start_1 = "movl      $111, %ebx # IACA START\n"
        mov_start_2 = "mov      $111, %ebx # IACA START\n"
        mov_end_1 = "movl      $222, %ebx # IACA END\n"
        mov_end_2 = "mov      $222, %ebx # IACA END\n"
        prologue = "movl    -92(%rbp), %r11d\n" + "movl      $111, %ebx\n"
        kernel = (
            "vfmadd132sd (%r15,%rcx,8), %xmm5, %xmm0\n"
            + "vmovsd  %xmm0, (%r14,%rcx,8)\n"
            + "cmpl    %ebx, %ecx\n"
            + "jge .L8\n"
        )
        epilogue = ".LE9:\t\t#12.2\n" "call    dummy\n"
        kernel_length = len(list(filter(None, kernel.split("\n"))))

        bytes_variations = [
            bytes_1_line,
            bytes_2_lines_1,
            bytes_2_lines_2,
            bytes_3_lines,
        ]
        mov_start_variations = [mov_start_1, mov_start_2]
        mov_end_variations = [mov_end_1, mov_end_2]
        # actual tests
        for mov_start_var in mov_start_variations:
            for bytes_var_1 in bytes_variations:
                for mov_end_var in mov_end_variations:
                    for bytes_var_2 in bytes_variations:
                        sample_code = (
                            prologue
                            + mov_start_var
                            + bytes_var_1
                            + kernel
                            + mov_end_var
                            + bytes_var_2
                            + epilogue
                        )
                        with self.subTest(
                            mov_start=mov_start_var,
                            bytes_start=bytes_var_1,
                            mov_end=mov_end_var,
                            bytes_end=bytes_var_2,
                        ):
                            sample_parsed = self.parser_x86.parse_file(sample_code)
                            sample_kernel = reduce_to_section(sample_parsed, "x86")
                            self.assertEqual(len(sample_kernel), kernel_length)
                            kernel_start = len(
                                list(
                                    filter(
                                        None,
                                        (prologue + mov_start_var + bytes_var_1).split("\n"),
                                    )
                                )
                            )
                            parsed_kernel = self.parser_x86.parse_file(
                                kernel, start_line=kernel_start
                            )
                            self.assertEqual(sample_kernel, parsed_kernel)

    def test_marker_special_cases_AArch(self):
        bytes_line = ".byte     213,3,32,31\n"
        start_marker = "mov      x1, #111\n" + bytes_line
        end_marker = "mov      x1, #222\n" + bytes_line
        prologue = "dup v0.2d, x14\n" "neg x9, x9\n" ".p2align    6\n"
        kernel = (
            ".LBB0_28:\n"
            + "fmul    v7.2d, v7.2d, v19.2d\n"
            + "stp q0, q1, [x10, #-32]\n"
            + "b.ne    .LBB0_28\n"
        )
        epilogue = ".LBB0_29:   //   Parent Loop BB0_20 Depth=1\n" "bl    dummy\n"

        samples = [
            # (test name,
            #  ignored prologue, section to be extraced, ignored epilogue)
            ("markers", prologue + start_marker, kernel, end_marker + epilogue),
            ("marker at file start", start_marker, kernel, end_marker + epilogue),
            ("no start marker", "", prologue + kernel, end_marker + epilogue),
            ("marker at file end", prologue + start_marker, kernel, end_marker),
            ("no end marker", prologue + start_marker, kernel + epilogue, ""),
            ("empty kernel", prologue + start_marker, "", end_marker + epilogue),
        ]

        for test_name, pro, kernel, epi in samples:
            code = pro + kernel + epi
            parsed = self.parser_AArch.parse_file(code)
            test_kernel = reduce_to_section(parsed, "AArch64")
            if kernel:
                kernel_length = len(kernel.strip().split("\n"))
            else:
                kernel_length = 0
            self.assertEqual(
                len(test_kernel),
                kernel_length,
                msg="Invalid exctracted kernel length on {!r} sample".format(test_name),
            )
            if pro:
                kernel_start = len((pro).strip().split("\n"))
            else:
                kernel_start = 0
            parsed_kernel = self.parser_AArch.parse_file(kernel, start_line=kernel_start)
            self.assertEqual(
                test_kernel,
                parsed_kernel,
                msg="Invalid exctracted kernel on {!r}".format(test_name),
            )

    def test_marker_special_cases_x86(self):
        bytes_line = ".byte     100\n" ".byte     103\n" ".byte     144\n"
        start_marker = "movl     $111, %ebx\n" + bytes_line
        end_marker = "movl     $222, %ebx\n" + bytes_line
        prologue = "movl    -88(%rbp), %r10d\n" "xorl    %r11d, %r11d\n" ".p2align 4,,10\n"
        kernel = (
            ".L3: #L3\n"
            "vmovsd  .LC1(%rip), %xmm0\n"
            "vmovsd  %xmm0, (%r15,%rcx,8)\n"
            "cmpl    %ecx, %ebx\n"
            "jle .L3\n"
        )
        epilogue = "leaq    -56(%rbp), %rsi\n" "movl    %r10d, -88(%rbp)\n" "call    timing\n"
        samples = [
            # (test name,
            #  ignored prologue, section to be extraced, ignored epilogue)
            ("markers", prologue + start_marker, kernel, end_marker + epilogue),
            ("marker at file start", start_marker, kernel, end_marker + epilogue),
            ("no start marker", "", prologue + kernel, end_marker + epilogue),
            ("marker at file end", prologue + start_marker, kernel, end_marker),
            ("no end marker", prologue + start_marker, kernel + epilogue, ""),
            ("empty kernel", prologue + start_marker, "", end_marker + epilogue),
        ]

        for test_name, pro, kernel, epi in samples:
            code = pro + kernel + epi
            parsed = self.parser_x86.parse_file(code)
            test_kernel = reduce_to_section(parsed, "x86")
            if kernel:
                kernel_length = len(kernel.strip().split("\n"))
            else:
                kernel_length = 0
            self.assertEqual(
                len(test_kernel),
                kernel_length,
                msg="Invalid exctracted kernel length on {!r} sample".format(test_name),
            )
            if pro:
                kernel_start = len((pro).strip().split("\n"))
            else:
                kernel_start = 0
            parsed_kernel = self.parser_x86.parse_file(kernel, start_line=kernel_start)
            self.assertEqual(
                test_kernel,
                parsed_kernel,
                msg="Invalid exctracted kernel on {!r}".format(test_name),
            )

    def test_find_jump_labels(self):
        self.assertEqual(
            find_jump_labels(self.parsed_x86),
            OrderedDict(
                [
                    (".LFB24", 10),
                    (".L4", 65),
                    (".L3", 79),
                    (".L2", 102),
                    (".L13", 111),
                    (".L12", 120),
                    (".L6", 132),
                    (".L10", 145),
                    (".L9", 161),
                    (".L8", 183),
                    (".L15", 252),
                    (".L26", 256),
                    (".L14", 259),
                    (".LFB25", 277),
                    (".L28", 289),
                ]
            ),
        )

        self.assertEqual(
            find_jump_labels(self.parsed_AArch),
            OrderedDict(
                [
                    ("triad", 18),
                    (".LBB0_3", 71),
                    (".LBB0_4", 76),
                    (".LBB0_5", 84),
                    (".LBB0_7", 92),
                    (".LBB0_8", 95),
                    (".LBB0_9", 106),
                    (".LBB0_11", 118),
                    (".LBB0_12", 133),
                    (".LBB0_14", 177),
                    (".LBB0_15", 190),
                    (".LBB0_16", 205),
                    (".LBB0_17", 208),
                    (".LBB0_18", 221),
                    (".LBB0_19", 228),
                    (".LBB0_20", 260),
                    (".LBB0_22", 272),
                    (".LBB0_24", 283),
                    (".LBB0_26", 290),
                    (".LBB0_28", 298),
                    (".LBB0_29", 306),
                    (".LBB0_31", 448),
                    (".LBB0_32", 458),
                    (".LBB0_33", 480),
                    (".LBB0_34", 484),
                    (".LBB0_35", 493),
                    (".LBB0_36", 504),
                    (".LBB0_37", 508),
                    (".LBB0_38", 518),
                    ("main", 574),
                ]
            ),
        )

    def test_find_basic_blocks(self):
        self.assertEqual(
            [
                (k, v[0]["line_number"], v[-1]["line_number"])
                for k, v in find_basic_blocks(self.parsed_x86).items()
            ],
            [
                (".LFB24", 11, 56),
                (".L4", 66, 74),
                (".L3", 80, 89),
                (".L2", 103, 112),
                (".L13", 112, 121),
                (".L12", 121, 125),
                (".L6", 133, 135),
                (".L10", 146, 154),
                (".L9", 162, 170),
                (".L8", 184, 187),
                (".L15", 253, 256),
                (".L26", 257, 259),
                (".L14", 260, 262),
                (".LFB25", 278, 290),
                (".L28", 290, 300),
            ],
        )

        self.assertEqual(
            [
                (k, v[0]["line_number"], v[-1]["line_number"])
                for k, v in find_basic_blocks(self.parsed_AArch).items()
            ],
            [
                ("triad", 19, 64),
                (".LBB0_3", 72, 77),
                (".LBB0_4", 77, 83),
                (".LBB0_5", 85, 89),
                (".LBB0_7", 93, 95),
                (".LBB0_8", 96, 105),
                (".LBB0_9", 107, 114),
                (".LBB0_11", 119, 134),
                (".LBB0_12", 134, 173),
                (".LBB0_14", 178, 191),
                (".LBB0_15", 191, 205),
                (".LBB0_16", 206, 208),
                (".LBB0_17", 209, 222),
                (".LBB0_18", 222, 228),
                (".LBB0_19", 229, 261),
                (".LBB0_20", 261, 269),
                (".LBB0_22", 273, 280),
                (".LBB0_24", 284, 286),
                (".LBB0_26", 291, 293),
                (".LBB0_28", 299, 307),
                (".LBB0_29", 307, 444),
                (".LBB0_31", 449, 459),
                (".LBB0_32", 459, 480),
                (".LBB0_33", 481, 484),
                (".LBB0_34", 485, 494),
                (".LBB0_35", 494, 504),
                (".LBB0_36", 505, 508),
                (".LBB0_37", 509, 518),
                (".LBB0_38", 519, 568),
                ("main", 575, 590),
            ],
        )

    def test_find_basic_loop_body(self):
        self.assertEqual(
            [
                (k, v[0]["line_number"], v[-1]["line_number"])
                for k, v in find_basic_loop_bodies(self.parsed_x86).items()
            ],
            [(".L4", 66, 74), (".L10", 146, 154), (".L28", 290, 300)],
        )

        self.assertEqual(
            [
                (k, v[0]["line_number"], v[-1]["line_number"])
                for k, v in find_basic_loop_bodies(self.parsed_AArch).items()
            ],
            [
                (".LBB0_12", 134, 173),
                (".LBB0_15", 191, 205),
                (".LBB0_18", 222, 228),
                (".LBB0_29", 307, 444),
                (".LBB0_32", 459, 480),
                (".LBB0_35", 494, 504),
            ],
        )

    ##################
    # Helper functions
    ##################

    @staticmethod
    def _find_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestMarkerUtils)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python3
"""
Unit tests for OSACA Frontend
"""

import os
import unittest

from osaca.frontend import Frontend
from osaca.parser import ParserAArch64, ParserX86ATT
from osaca.semantics import ArchSemantics, KernelDG, MachineModel


class TestFrontend(unittest.TestCase):
    MODULE_DATA_DIR = os.path.join(
        os.path.dirname(os.path.split(os.path.abspath(__file__))[0]), "osaca/data/"
    )

    @classmethod
    def setUpClass(self):
        # set up parser and kernels
        self.parser_x86 = ParserX86ATT()
        self.parser_AArch64 = ParserAArch64()
        with open(self._find_file("kernel_x86.s")) as f:
            code_x86 = f.read()
        with open(self._find_file("kernel_aarch64.s")) as f:
            code_AArch64 = f.read()
        self.kernel_x86 = self.parser_x86.parse_file(code_x86)
        self.kernel_AArch64 = self.parser_AArch64.parse_file(code_AArch64)

        # set up machine models
        self.machine_model_csx = MachineModel(
            path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "csx.yml")
        )
        self.machine_model_tx2 = MachineModel(arch="tx2")
        self.semantics_csx = ArchSemantics(
            self.machine_model_csx,
            path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "isa/x86.yml"),
        )
        self.semantics_tx2 = ArchSemantics(
            self.machine_model_tx2,
            path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "isa/aarch64.yml"),
        )
        for i in range(len(self.kernel_x86)):
            self.semantics_csx.assign_src_dst(self.kernel_x86[i])
            self.semantics_csx.assign_tp_lt(self.kernel_x86[i])
        for i in range(len(self.kernel_AArch64)):
            self.semantics_tx2.assign_src_dst(self.kernel_AArch64[i])
            self.semantics_tx2.assign_tp_lt(self.kernel_AArch64[i])

    ###########
    # Tests
    ###########

    def test_frontend_creation(self):
        with self.assertRaises(ValueError):
            Frontend()
        with self.assertRaises(ValueError):
            Frontend(arch="csx", path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "csx.yml"))
        with self.assertRaises(FileNotFoundError):
            Frontend(path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "THE_MACHINE.yml"))
        with self.assertRaises(FileNotFoundError):
            Frontend(arch="THE_MACHINE")
        Frontend(arch="zen1")

    def test_frontend_x86(self):
        dg = KernelDG(self.kernel_x86, self.parser_x86, self.machine_model_csx, self.semantics_csx)
        fe = Frontend(path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "csx.yml"))
        fe.throughput_analysis(self.kernel_x86, show_cmnts=False)
        fe.latency_analysis(dg.get_critical_path())
        # TODO compare output with checked string

    def test_frontend_AArch64(self):
        dg = KernelDG(
            self.kernel_AArch64,
            self.parser_AArch64,
            self.machine_model_tx2,
            self.semantics_tx2,
        )
        fe = Frontend(path_to_yaml=os.path.join(self.MODULE_DATA_DIR, "tx2.yml"))
        fe.full_analysis(self.kernel_AArch64, dg, verbose=True)
        # TODO compare output with checked string

    ##################
    # Helper functions
    ##################

    @staticmethod
    def _find_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFrontend)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python3

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from setuptools.command.install import install as _install
from setuptools.command.sdist import sdist as _sdist

# To use a consistent encoding
from codecs import open
import os
import io
import re
import sys

here = os.path.abspath(os.path.dirname(__file__))


# Stolen from pip
def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8"),
    ) as fp:
        return fp.read()


# Stolen from pip
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def _run_build_cache(dir):
    from subprocess import check_call

    # This is run inside the install staging directory (that had no .pyc files)
    # We don't want to generate any.
    # https://github.com/eliben/pycparser/pull/135
    check_call(
        [sys.executable, "-B", "_build_cache.py"],
        cwd=os.path.join(dir, "osaca", "data"),
    )


class install(_install):
    def run(self):
        _install.run(self)
        self.execute(
            _run_build_cache,
            (self.install_lib,),
            msg="Build ISA and architecture cache",
        )


class sdist(_sdist):
    def make_release_tree(self, basedir, files):
        _sdist.make_release_tree(self, basedir, files)
        self.execute(_run_build_cache, (basedir,), msg="Build ISA and architecture cache")


# Get the long description from the README file
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="osaca",
    # Version should comply with PEP440. For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    #  https://packaging.python.org/en/latest/distributing.html
    version=find_version("osaca", "__init__.py"),
    description="Open Source Architecture Code Analyzer",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # The project's main homepage
    url="https://github.com/RRZE-HPC/OSACA",
    # Author details
    author="Jan Laukemann",
    author_email="jan.laukemann@fau.de",
    # Choose your license
    license="AGPLv3",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Utilities",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: GNU Affero General Public License v3",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate wheter you support Python2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    # What doesd your project relate to?
    keywords="hpc performance benchmark analysis architecture",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["networkx", "pyparsing>=2.3.1", "ruamel.yaml>=0.15.71"],
    python_requires=">=3.6",
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    # },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data=True,
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={"console_scripts": ["osaca=osaca.osaca:main"]},
    # Overwriting install and sdist to enforce cache distribution with package
    cmdclass={"install": install, "sdist": sdist},
)



#!/usr/bin/env python3
from glob import glob
import os.path
import sys

sys.path[0:0] = ["../.."]

failed = False
try:
    from osaca.semantics.hw_model import MachineModel
except ModuleNotFoundError:
    print(
        "Unable to import MachineModel, probably some dependency is not yet installed. SKIPPING. "
        "First run of OSACA may take a while to build caches, subsequent runs will be as fast as "
        "ever."
    )
    sys.exit()

print("Building cache: ", end="")
sys.stdout.flush()

# Iterating architectures
for f in glob(os.path.join(os.path.dirname(__file__), "*.yml")):
    MachineModel(path_to_yaml=f)
    print(".", end="")
    sys.stdout.flush()

# Iterating ISAs
for f in glob(os.path.join(os.path.dirname(__file__), "isa/*.yml")):
    MachineModel(path_to_yaml=f)
    print("+", end="")
    sys.stdout.flush()

print()



"""Open Source Architecture Code Analyzer"""
name = "osaca"
__version__ = "0.4.6"

# To trigger travis deployment to pypi, do the following:
# 1. Increment __version___
# 2. commit to RRZE-HPC/osaca's master branch
# 3. wait for Github Actions to complete successful (unless already tested)
# 4. tag commit with 'v{}'.format(__version__) (`git tag vX.Y.Z`)
# 5. push tag to github (`git push origin vX.Y.Z` or push all tags with `git push --tags`)



#!/usr/bin/env python3
"""
Unit tests for ARMv8 AArch64 assembly parser
"""

import os
import unittest

from pyparsing import ParseException

from osaca.parser import AttrDict, ParserAArch64


class TestParserAArch64(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.parser = ParserAArch64()
        with open(self._find_file("triad_arm_iaca.s")) as f:
            self.triad_code = f.read()

    ##################
    # Test
    ##################

    def test_comment_parser(self):
        self.assertEqual(self._get_comment(self.parser, "// some comments"), "some comments")
        self.assertEqual(
            self._get_comment(self.parser, "\t\t//AA BB CC \t end \t"), "AA BB CC end"
        )
        self.assertEqual(
            self._get_comment(self.parser, "\t//// comment //// comment"),
            "// comment //// comment",
        )

    def test_label_parser(self):
        self.assertEqual(self._get_label(self.parser, "main:").name, "main")
        self.assertEqual(self._get_label(self.parser, "..B1.10:").name, "..B1.10")
        self.assertEqual(self._get_label(self.parser, ".2.3_2_pack.3:").name, ".2.3_2_pack.3")
        self.assertEqual(self._get_label(self.parser, ".L1:\t\t\t//label1").name, ".L1")
        self.assertEqual(
            " ".join(self._get_label(self.parser, ".L1:\t\t\t//label1").comment),
            "label1",
        )
        with self.assertRaises(ParseException):
            self._get_label(self.parser, "\t.cfi_startproc")

    def test_directive_parser(self):
        self.assertEqual(self._get_directive(self.parser, "\t.text").name, "text")
        self.assertEqual(len(self._get_directive(self.parser, "\t.text").parameters), 0)
        self.assertEqual(self._get_directive(self.parser, "\t.align\t16,0x90").name, "align")
        self.assertEqual(len(self._get_directive(self.parser, "\t.align\t16,0x90").parameters), 2)
        self.assertEqual(
            self._get_directive(self.parser, "\t.align\t16,0x90").parameters[1], "0x90"
        )
        self.assertEqual(
            self._get_directive(self.parser, "        .byte 100,103,144       //IACA START")[
                "name"
            ],
            "byte",
        )
        self.assertEqual(
            self._get_directive(self.parser, "        .byte 100,103,144       //IACA START")[
                "parameters"
            ][2],
            "144",
        )
        self.assertEqual(
            " ".join(
                self._get_directive(self.parser, "        .byte 100,103,144       //IACA START")[
                    "comment"
                ]
            ),
            "IACA START",
        )

    def test_parse_instruction(self):
        instr1 = "\t\tvcvt.F32.S32 w1, w2\t\t\t//12.27"
        instr2 = "b.lo        ..B1.4 \t"
        instr3 = "        mov x2,#0x222          //NOT IACA END"
        instr4 = "str x28, [sp, x1, lsl #4] //12.9"
        instr5 = "ldr x0, [x0, #:got_lo12:q2c]"
        instr6 = "adrp    x0, :got:visited"
        instr7 = "fadd    v17.2d, v16.2d, v1.2d"

        parsed_1 = self.parser.parse_instruction(instr1)
        parsed_2 = self.parser.parse_instruction(instr2)
        parsed_3 = self.parser.parse_instruction(instr3)
        parsed_4 = self.parser.parse_instruction(instr4)
        parsed_5 = self.parser.parse_instruction(instr5)
        parsed_6 = self.parser.parse_instruction(instr6)
        parsed_7 = self.parser.parse_instruction(instr7)

        self.assertEqual(parsed_1.instruction, "vcvt.F32.S32")
        self.assertEqual(parsed_1.operands[0].register.name, "1")
        self.assertEqual(parsed_1.operands[0].register.prefix, "w")
        self.assertEqual(parsed_1.operands[1].register.name, "2")
        self.assertEqual(parsed_1.operands[1].register.prefix, "w")
        self.assertEqual(parsed_1.comment, "12.27")

        self.assertEqual(parsed_2.instruction, "b.lo")
        self.assertEqual(parsed_2.operands[0].identifier.name, "..B1.4")
        self.assertEqual(len(parsed_2.operands), 1)
        self.assertIsNone(parsed_2.comment)

        self.assertEqual(parsed_3.instruction, "mov")
        self.assertEqual(parsed_3.operands[0].register.name, "2")
        self.assertEqual(parsed_3.operands[0].register.prefix, "x")
        self.assertEqual(parsed_3.operands[1].immediate.value, int("0x222", 0))
        self.assertEqual(parsed_3.comment, "NOT IACA END")

        self.assertEqual(parsed_4.instruction, "str")
        self.assertIsNone(parsed_4.operands[1].memory.offset)
        self.assertEqual(parsed_4.operands[1].memory.base.name, "sp")
        self.assertEqual(parsed_4.operands[1].memory.base.prefix, "x")
        self.assertEqual(parsed_4.operands[1].memory.index.name, "1")
        self.assertEqual(parsed_4.operands[1].memory.index.prefix, "x")
        self.assertEqual(parsed_4.operands[1].memory.scale, 16)
        self.assertEqual(parsed_4.operands[0].register.name, "28")
        self.assertEqual(parsed_4.operands[0].register.prefix, "x")
        self.assertEqual(parsed_4.comment, "12.9")

        self.assertEqual(parsed_5.instruction, "ldr")
        self.assertEqual(parsed_5.operands[0].register.name, "0")
        self.assertEqual(parsed_5.operands[0].register.prefix, "x")
        self.assertEqual(parsed_5.operands[1].memory.offset.identifier.name, "q2c")
        self.assertEqual(parsed_5.operands[1].memory.offset.identifier.relocation, ":got_lo12:")
        self.assertEqual(parsed_5.operands[1].memory.base.name, "0")
        self.assertEqual(parsed_5.operands[1].memory.base.prefix, "x")
        self.assertIsNone(parsed_5.operands[1].memory.index)
        self.assertEqual(parsed_5.operands[1].memory.scale, 1)

        self.assertEqual(parsed_6.instruction, "adrp")
        self.assertEqual(parsed_6.operands[0].register.name, "0")
        self.assertEqual(parsed_6.operands[0].register.prefix, "x")
        self.assertEqual(parsed_6.operands[1].identifier.relocation, ":got:")
        self.assertEqual(parsed_6.operands[1].identifier.name, "visited")

        self.assertEqual(parsed_7.instruction, "fadd")
        self.assertEqual(parsed_7.operands[0].register.name, "17")
        self.assertEqual(parsed_7.operands[0].register.prefix, "v")
        self.assertEqual(parsed_7.operands[0].register.lanes, "2")
        self.assertEqual(parsed_7.operands[0].register.shape, "d")
        self.assertEqual(self.parser.get_full_reg_name(parsed_7.operands[2].register), "v1.2d")

    def test_parse_line(self):
        line_comment = "// -- Begin  main"
        line_label = ".LBB0_1:              // =>This Inner Loop Header: Depth=1"
        line_directive = ".cfi_def_cfa w29, -16"
        line_instruction = "ldr s0, [x11, w10, sxtw #2]    // = <<2"
        line_prefetch = "prfm    pldl1keep, [x26, #2048] //HPL"
        line_preindexed = "stp x29, x30, [sp, #-16]!"
        line_postindexed = "ldp q2, q3, [x11], #64"
        line_5_operands = "fcmla z26.d, p0/m, z29.d, z21.d, #90"

        instruction_form_1 = {
            "instruction": None,
            "operands": [],
            "directive": None,
            "comment": "-- Begin main",
            "label": None,
            "line": "// -- Begin  main",
            "line_number": 1,
        }

        instruction_form_2 = {
            "instruction": None,
            "operands": [],
            "directive": None,
            "comment": "=>This Inner Loop Header: Depth=1",
            "label": ".LBB0_1",
            "line": ".LBB0_1:              // =>This Inner Loop Header: Depth=1",
            "line_number": 2,
        }
        instruction_form_3 = {
            "instruction": None,
            "operands": [],
            "directive": {"name": "cfi_def_cfa", "parameters": ["w29", "-16"]},
            "comment": None,
            "label": None,
            "line": ".cfi_def_cfa w29, -16",
            "line_number": 3,
        }
        instruction_form_4 = {
            "instruction": "ldr",
            "operands": [
                {"register": {"prefix": "s", "name": "0"}},
                {
                    "memory": {
                        "offset": None,
                        "base": {"prefix": "x", "name": "11"},
                        "index": {
                            "prefix": "w",
                            "name": "10",
                            "shift_op": "sxtw",
                            "immediate": {"value": "2"},
                            "shift": [{"value": "2"}],
                        },
                        "scale": 4,
                    }
                },
            ],
            "directive": None,
            "comment": "= <<2",
            "label": None,
            "line": "ldr s0, [x11, w10, sxtw #2]    // = <<2",
            "line_number": 4,
        }
        instruction_form_5 = {
            "instruction": "prfm",
            "operands": [
                {"prfop": {"type": ["PLD"], "target": ["L1"], "policy": ["KEEP"]}},
                {
                    "memory": {
                        "offset": {"value": 2048},
                        "base": {"prefix": "x", "name": "26"},
                        "index": None,
                        "scale": 1,
                    }
                },
            ],
            "directive": None,
            "comment": "HPL",
            "label": None,
            "line": "prfm    pldl1keep, [x26, #2048] //HPL",
            "line_number": 5,
        }
        instruction_form_6 = {
            "instruction": "stp",
            "operands": [
                {"register": {"prefix": "x", "name": "29"}},
                {"register": {"prefix": "x", "name": "30"}},
                {
                    "memory": {
                        "offset": {"value": -16},
                        "base": {"name": "sp", "prefix": "x"},
                        "index": None,
                        "scale": 1,
                        "pre_indexed": True,
                    }
                },
            ],
            "directive": None,
            "comment": None,
            "label": None,
            "line": "stp x29, x30, [sp, #-16]!",
            "line_number": 6,
        }
        instruction_form_7 = {
            "instruction": "ldp",
            "operands": [
                {"register": {"prefix": "q", "name": "2"}},
                {"register": {"prefix": "q", "name": "3"}},
                {
                    "memory": {
                        "offset": None,
                        "base": {"prefix": "x", "name": "11"},
                        "index": None,
                        "scale": 1,
                        "post_indexed": {"value": 64},
                    }
                },
            ],
            "directive": None,
            "comment": None,
            "label": None,
            "line": "ldp q2, q3, [x11], #64",
            "line_number": 7,
        }
        instruction_form_8 = {
            "instruction": "fcmla",
            "operands": [
                {"register": {"prefix": "z", "name": "26", "shape": "d"}},
                {"register": {"prefix": "p", "name": "0", "predication": "m"}},
                {"register": {"prefix": "z", "name": "29", "shape": "d"}},
                {"register": {"prefix": "z", "name": "21", "shape": "d"}},
                {"immediate": {"value": 90, "type": "int"}},
            ],
            "directive": None,
            "comment": None,
            "label": None,
            "line": "fcmla z26.d, p0/m, z29.d, z21.d, #90",
            "line_number": 8,
        }

        parsed_1 = self.parser.parse_line(line_comment, 1)
        parsed_2 = self.parser.parse_line(line_label, 2)
        parsed_3 = self.parser.parse_line(line_directive, 3)
        parsed_4 = self.parser.parse_line(line_instruction, 4)
        parsed_5 = self.parser.parse_line(line_prefetch, 5)
        parsed_6 = self.parser.parse_line(line_preindexed, 6)
        parsed_7 = self.parser.parse_line(line_postindexed, 7)
        parsed_8 = self.parser.parse_line(line_5_operands, 8)

        self.assertEqual(parsed_1, instruction_form_1)
        self.assertEqual(parsed_2, instruction_form_2)
        self.assertEqual(parsed_3, instruction_form_3)
        self.assertEqual(parsed_4, instruction_form_4)
        self.assertEqual(parsed_5, instruction_form_5)
        self.assertEqual(parsed_6, instruction_form_6)
        self.assertEqual(parsed_7, instruction_form_7)
        self.assertEqual(parsed_8, instruction_form_8)

    def test_parse_file(self):
        parsed = self.parser.parse_file(self.triad_code)
        self.assertEqual(parsed[0].line_number, 1)
        self.assertEqual(len(parsed), 645)

    def test_normalize_imd(self):
        imd_decimal_1 = {"value": "79"}
        imd_hex_1 = {"value": "0x4f"}
        imd_decimal_2 = {"value": "8"}
        imd_hex_2 = {"value": "0x8"}
        imd_float_11 = {"float": {"mantissa": "0.79", "e_sign": "+", "exponent": "2"}}
        imd_float_12 = {"float": {"mantissa": "790.0", "e_sign": "-", "exponent": "1"}}
        imd_double_11 = {"double": {"mantissa": "0.79", "e_sign": "+", "exponent": "2"}}
        imd_double_12 = {"double": {"mantissa": "790.0", "e_sign": "-", "exponent": "1"}}
        identifier = {"identifier": {"name": "..B1.4"}}

        value1 = self.parser.normalize_imd(imd_decimal_1)
        self.assertEqual(value1, self.parser.normalize_imd(imd_hex_1))
        self.assertEqual(
            self.parser.normalize_imd(imd_decimal_2),
            self.parser.normalize_imd(imd_hex_2),
        )
        self.assertEqual(self.parser.normalize_imd(imd_float_11), value1)
        self.assertEqual(self.parser.normalize_imd(imd_float_12), value1)
        self.assertEqual(self.parser.normalize_imd(imd_double_11), value1)
        self.assertEqual(self.parser.normalize_imd(imd_double_12), value1)
        self.assertEqual(self.parser.normalize_imd(identifier), identifier)

    def test_multiple_regs(self):
        instr_range = "PUSH {x5-x7}"
        instr_list = "POP {x5, x6, x7}"
        instr_range_with_index = "ld4 {v0.S - v3.S}[2]"
        instr_list_with_index = "ld4 {v0.S, v1.S, v2.S, v3.S}[2]"
        instr_range_single = "dummy  { z1.d }"
        reg_list = [
            AttrDict({"register": {"prefix": "x", "name": "5"}}),
            AttrDict({"register": {"prefix": "x", "name": "6"}}),
            AttrDict({"register": {"prefix": "x", "name": "7"}}),
        ]
        reg_list_idx = [
            AttrDict({"register": {"prefix": "v", "name": "0", "shape": "S", "index": 2}}),
            AttrDict({"register": {"prefix": "v", "name": "1", "shape": "S", "index": 2}}),
            AttrDict({"register": {"prefix": "v", "name": "2", "shape": "S", "index": 2}}),
            AttrDict({"register": {"prefix": "v", "name": "3", "shape": "S", "index": 2}}),
        ]
        reg_list_single = [AttrDict({"register": {"prefix": "z", "name": "1", "shape": "d"}})]

        prange = self.parser.parse_line(instr_range)
        plist = self.parser.parse_line(instr_list)
        p_idx_range = self.parser.parse_line(instr_range_with_index)
        p_idx_list = self.parser.parse_line(instr_list_with_index)
        p_single = self.parser.parse_line(instr_range_single)

        self.assertEqual(prange.operands, reg_list)
        self.assertEqual(plist.operands, reg_list)
        self.assertEqual(p_idx_range.operands, reg_list_idx)
        self.assertEqual(p_idx_list.operands, reg_list_idx)
        self.assertEqual(p_single.operands, reg_list_single)

    def test_reg_dependency(self):
        reg_1_1 = AttrDict({"prefix": "b", "name": "1"})
        reg_1_2 = AttrDict({"prefix": "h", "name": "1"})
        reg_1_3 = AttrDict({"prefix": "s", "name": "1"})
        reg_1_4 = AttrDict({"prefix": "d", "name": "1"})
        reg_1_4 = AttrDict({"prefix": "q", "name": "1"})
        reg_2_1 = AttrDict({"prefix": "w", "name": "2"})
        reg_2_2 = AttrDict({"prefix": "x", "name": "2"})
        reg_v1_1 = AttrDict({"prefix": "v", "name": "11", "lanes": "16", "shape": "b"})
        reg_v1_2 = AttrDict({"prefix": "v", "name": "11", "lanes": "8", "shape": "h"})
        reg_v1_3 = AttrDict({"prefix": "v", "name": "11", "lanes": "4", "shape": "s"})
        reg_v1_4 = AttrDict({"prefix": "v", "name": "11", "lanes": "2", "shape": "d"})

        reg_b5 = AttrDict({"prefix": "b", "name": "5"})
        reg_q15 = AttrDict({"prefix": "q", "name": "15"})
        reg_v10 = AttrDict({"prefix": "v", "name": "10", "lanes": "2", "shape": "s"})
        reg_v20 = AttrDict({"prefix": "v", "name": "20", "lanes": "2", "shape": "d"})

        reg_1 = [reg_1_1, reg_1_2, reg_1_3, reg_1_4]
        reg_2 = [reg_2_1, reg_2_2]
        reg_v = [reg_v1_1, reg_v1_2, reg_v1_3, reg_v1_4]
        reg_others = [reg_b5, reg_q15, reg_v10, reg_v20]
        regs = reg_1 + reg_2 + reg_v + reg_others

        # test each register against each other
        for ri in reg_1:
            for rj in regs:
                assert_value = True if rj in reg_1 else False
                with self.subTest(reg_a=ri, reg_b=rj, assert_val=assert_value):
                    self.assertEqual(self.parser.is_reg_dependend_of(ri, rj), assert_value)
        for ri in reg_2:
            for rj in regs:
                assert_value = True if rj in reg_2 else False
                with self.subTest(reg_a=ri, reg_b=rj, assert_val=assert_value):
                    self.assertEqual(self.parser.is_reg_dependend_of(ri, rj), assert_value)
        for ri in reg_v:
            for rj in regs:
                assert_value = True if rj in reg_v else False
                with self.subTest(reg_a=ri, reg_b=rj, assert_val=assert_value):
                    self.assertEqual(self.parser.is_reg_dependend_of(ri, rj), assert_value)
        for ri in reg_others:
            for rj in regs:
                assert_value = True if rj == ri else False
                with self.subTest(reg_a=ri, reg_b=rj, assert_val=assert_value):
                    self.assertEqual(self.parser.is_reg_dependend_of(ri, rj), assert_value)

    ##################
    # Helper functions
    ##################
    def _get_comment(self, parser, comment):
        return " ".join(
            AttrDict.convert_dict(
                parser.process_operand(parser.comment.parseString(comment, parseAll=True).asDict())
            ).comment
        )

    def _get_label(self, parser, label):
        return AttrDict.convert_dict(
            parser.process_operand(parser.label.parseString(label, parseAll=True).asDict())
        ).label

    def _get_directive(self, parser, directive):
        return AttrDict.convert_dict(
            parser.process_operand(parser.directive.parseString(directive, parseAll=True).asDict())
        ).directive

    @staticmethod
    def _find_file(name):
        testdir = os.path.dirname(__file__)
        name = os.path.join(testdir, "test_files", name)
        assert os.path.exists(name)
        return name


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestParserAArch64)
    unittest.TextTestRunner(verbosity=2).run(suite)
